import pandas as pd
from frozendict import frozendict as fdict
import functools


def get_hex_string(c):
    return '#{:02X}{:02X}{:02X}'.format(*c)


def cluster_color(v=None):
    phylogic_color_list = [[166, 17, 129],
                           [39, 140, 24],
                           [103, 200, 243],
                           [248, 139, 16],
                           [16, 49, 41],
                           [93, 119, 254],
                           [152, 22, 26],
                           [104, 236, 172],
                           [249, 142, 135],
                           [55, 18, 48],
                           [83, 82, 22],
                           [247, 36, 36],
                           [0, 79, 114],
                           [243, 65, 132],
                           [60, 185, 179],
                           [185, 177, 243],
                           [139, 34, 67],
                           [178, 41, 186],
                           [58, 146, 231],
                           [130, 159, 21],
                           [161, 91, 243],
                           [131, 61, 17],
                           [248, 75, 81],
                           [32, 75, 32],
                           [45, 109, 116],
                           [255, 169, 199],
                           [55, 179, 113],
                           [34, 42, 3],
                           [56, 121, 166],
                           [172, 60, 15],
                           [115, 76, 204],
                           [21, 61, 73],
                           [67, 21, 74],  # Additional colors, uglier and bad
                           [123, 88, 112],
                           [87, 106, 46],
                           [37, 66, 58],
                           [132, 79, 62],
                           [71, 58, 32],
                           [59, 104, 114],
                           [46, 107, 90],
                           [84, 68, 73],
                           [90, 97, 124],
                           [121, 66, 76],
                           [104, 93, 48],
                           [49, 67, 82],
                           [71, 95, 65],
                           [127, 85, 44],  # even more additional colors, gray
                           [88, 79, 92],
                           [220, 212, 194],
                           [35, 34, 36],
                           [200, 220, 224],
                           [73, 81, 69],
                           [224, 199, 206],
                           [120, 127, 113],
                           [142, 148, 166],
                           [153, 167, 156],
                           [162, 139, 145],
                           [0, 0, 0]]  # black
    colors_dict = {str(i): get_hex_string(c) for i, c in enumerate(phylogic_color_list)}

    if v:
        return colors_dict[str(v)]
    else:
        return colors_dict


def get_unique_identifier(row, chrom='Chromosome', start_pos='Start_position',
                          ref='Reference_Allele', alt='Tumor_Seq_Allele'):
    """Generates unique string for this mutation, including contig, start position, ref and alt alleles.

    Does not include End Position, for this field is not present in mut_ccfs Phylogic output. However, specification of both the alt and ref alleles are enough to distinguish InDels.

    :param row: pd.Series giving the data for one mutation from a maf or maf-like dataframe
    :param chrom: the name of the contig/chromosome column/field; default: Chromosome
    :param start_pos: the name of the start position column/field; default: Start_position
    :param ref: the name of the reference allele column/field; default: Reference_Allele
    :param alt: the name of the alternate allele column/field; default: Tumor_Seq_Allele
    """
    return f"{row[chrom]}:{row[start_pos]}{row[ref]}>{row[alt]}"


@functools.lru_cache(maxsize=32)
def cached_read_csv(fn, **kwargs):
    """Convenience method: Pandas read_csv with a cache already implemented.

    Maxsize is set to 32 by default. At most, 32 function calls will be stored in memory.

    :param fn: filename/path
    :param kwargs: additional arguments to be passed to pandas.read_csv
    :return: pandas.DataFrame from given filename
    """
    return pd.read_csv(fn, **kwargs)


def freezeargs(func):
    """Transform mutable dictionary into immutable and lists into tuples

    Useful to be compatible with cache. Use: decorate functions (@freezeargs) that are decorated with a
    functools cache but have mutable inputs.

    :param func: function that is being wrapped
    """
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        args = tuple([fdict(arg) if isinstance(arg, dict) else (tuple(arg) if isinstance(arg, list) else arg) for arg in args])
        kwargs = {k: fdict(v) if isinstance(v, dict) else (tuple(v) if isinstance(v, list) else v) for k, v in kwargs.items()}
        return func(*args, **kwargs)
    return wrapped


from AnnoMate.Data import Data
import pandas as pd
from typing import Dict


class GenericData(Data):

    def __init__(self,
                 index,
                 description,
                 df: pd.DataFrame,
                 annot_df: pd.DataFrame = None,
                 annot_col_config_dict: Dict = None,
                 history_df: pd.DataFrame = None,
                 ):
        """
        A basic Data object with only one dataframe. This is useful if you are reviewing a
        single data type and all the data type (ie just sample data, just patient data, etc.) and/or
        all the data you want to review can be contained in a single dataframe.

        Parameters
        ----------
        df: pd.DataFrame
            A pandas dataframe. The index should correspond to the data to review.
        """
        super().__init__(index=index,
                         description=description,
                         annot_df=annot_df,
                         annot_col_config_dict=annot_col_config_dict,
                         history_df=history_df)

        self.df = df



import pandas as pd
from AnnoMate.Reviewers.ExampleReviewer import ExampleReviewer

def test_reviewer():
    fn = 'tutorial_notebooks/example_data/AnnoMate_Tutorial/data_to_review_example.tsv'
    df = pd.read_csv(fn, sep='\t')
    df = df.set_index('sample_id')
    assert df.shape[0] > 0

    my_reviewer = ExampleReviewer()
    output_pkl_path = './example_reviewer_data'
    my_reviewer.set_review_data(data_path=output_pkl_path, 
                                description='Intro to reviewers review session part 2',
                                sample_df=df,
                                preprocessing_str='Testing preprocessing')
    
    assert my_reviewer.review_data_interface.mh.metadata['freeze_data'] == False
    # test state of the reviewer on instantiation
    assert list(my_reviewer.list_data_attributes()) == ['index', 'description', 'annot_col_config_dict', 'annot_df', 'history_df', 'df']
    assert my_reviewer.get_data_attribute('df').equals(df)
    
    



from .ReviewDataInterface import ReviewDataInterface, DataAnnotation, Data
from .Data import validate_annot_data
from .ReviewDataApp import ReviewDataApp, valid_annotation_app_display_types, AnnotationDisplayComponent
from AnnoMate.AnnotationDisplayComponent import *
from AnnoMate.MetadataHandler import MetadataHandler
import pandas as pd
import os
from dash.dependencies import State
from typing import Union, Dict, List
from pathlib import Path
from abc import ABC, abstractmethod
import pathlib
import pickle
import inspect
import traceback
import warnings
from datetime import date
import functools
import gc # garbage collection


def make_docstring(object_type_name, func1_doc, func2):
    func2_doc = f"{object_type_name}.{func2.__name__}" + str(inspect.signature(func2)) + func2.__doc__
    if func2_doc not in func1_doc:
        return func1_doc + func2_doc
    else:
        return func1_doc


class ReviewerTemplate(ABC):
    
    def __init__(self):
        self.review_data_interface = None
        self.app = None
        self.autofill_dict = {}
        self.annot_app_display_types_dict = {}
        type(self).add_review_data_annotation.__doc__ = \
            make_docstring(object_type_name="ReviewDataInterface",
                           func1_doc=type(self).add_review_data_annotation.__doc__,
                           func2=ReviewDataInterface.add_annotation)

        try:
            type(self).set_review_app.__doc__ = \
                make_docstring(object_type_name=type(self).__name__,
                               func1_doc=type(self).set_review_app.__doc__,
                               func2=type(self).gen_review_app)

            type(self).set_review_data.__doc__ = \
                make_docstring(object_type_name=type(self).__name__,
                               func1_doc=type(self).set_review_data.__doc__,
                               func2=type(self).gen_data)
        except TypeError as e:
            raise TypeError(
                f"Docstrings may be missing from gen_review_app and/or gen_data.\n"
                f"Full trace:\n {traceback.format_exc()}")


    @abstractmethod
    def gen_data(self,
                 description: str,
                 annot_df: pd.DataFrame = None,
                 annot_col_config_dict: Dict = None,
                 history_df: pd.DataFrame = None,
                 index: List = None,
                 *args,
                 **kwargs) -> Data:
        """
        Specify type of data object to return and include additional kwargs
        for all the tables required for the Data Object type

        Returns
        -------
        Data
            Object that stores data to be reviewed and corresponding annotation and history

        """
        pass

    @abstractmethod
    def set_default_review_data_annotations(self):
        """Add annotations to review data object.
        Notes
        -----
        Call self.add_review_data_annotation() to set custom annotation configurations

        """
        pass

    @abstractmethod
    def gen_review_app(self, *args, **kwargs) -> ReviewDataApp:
        """
        Generates a ReviewDataApp object

        Returns
        -------
        ReviewDataApp
            App object that will display the dashboard

        Notes
        -----
        1. Instatiate a ReviewDataApp
        2. Add components

            ```
            app = ReviewDataApp()
            app.add_component(AppComponent(...))
            ```
        """
        pass

    @abstractmethod
    def set_default_review_data_annotations_app_display(self):
        """Define how annotation columns are displayed in the app.
        Notes
        -----
        Call self.add_review_data_annotations_app_display() to set custom display options

        """
        pass

    @abstractmethod
    def set_default_autofill(self):
        """Set which component values can be used for auto-filling input form

        Notes
        -----
        Call self.add_autofill() to set custom autofill options

        """
        pass
    
    # Public methods
    def set_review_data(self,
                        data_path: pathlib.Path,
                        description: str = None,
                        annot_df: pd.DataFrame = None,
                        annot_col_config_dict: pd.DataFrame = None,
                        history_df: pd.DataFrame = None,
                        load_existing_data_pkl_fn: Union[str, pathlib.Path] = None,
                        load_existing_exported_data_dir: Union[str, pathlib.Path] = None,
                        **kwargs):
        """Sets the review session ReviewData Object.

        Parameters
        ----------
        data_path : Union[str, Path]
            path to a directory where data pickle file and metadata 
            config file will be saved for current review session

        description : str,
            description of the data being reviewed

        annot_df : pd.DataFrame, optional
            dataframe of annotations from previous review sessions.

        annot_col_config_dict : dict(), optional
            dictionary defining the DataAnnotation configurations for the columns of annot_df

        history_df : pd.DataFrame, optional
            dataframe of history from previous review sessions.

        load_existing_data_pkl_fn : Union[str, Path], optional
            path to pickle file of a previous review session's data object.

        load_existing_exported_data_dir : Union[str, Path]
            path to a directory with exported annotation and history
            tables from a previous review session's data object

        **kwargs: dict
                  See additional parameters from self.gen_data() below

        """

        data_pkl_fn = f'{data_path}/data.pkl'
        metadata_config_fn = f'{data_path}/metadata_config.yaml'

        if not os.path.exists(data_path):
            os.makedirs(data_path)
        if not os.path.exists(metadata_config_fn) or not os.path.exists(data_pkl_fn):
            mh = MetadataHandler(metadata_config_fn, overwrite=True)
            mh.set_attribute('freeze_data', False)
            # can also add future defaults here 
        else:
            mh = MetadataHandler(metadata_config_fn)

        if os.path.exists(data_pkl_fn) and mh.metadata['freeze_data']:
            f = open(data_pkl_fn, 'rb')
            data = pickle.load(f)
        else:
            if description is None:
                raise ValueError(f'description is None. Provide a description if you are setting a new data object.')
            if (load_existing_data_pkl_fn is not None) and \
                    os.path.exists(load_existing_data_pkl_fn):
                print("Loading data from previous review with pickle file")
                f = open(load_existing_data_pkl_fn, 'rb')
                existing_data = pickle.load(f)
                f.close()

                annot_df = existing_data.annot_df
                annot_col_config_dict = existing_data.annot_col_config_dict
                history_df = existing_data.history_df

            elif (load_existing_exported_data_dir is not None) and \
                    os.path.exists(load_existing_exported_data_dir):
                print("Loading data from previous review with exported files")
                annot_df_fn = f'{load_existing_exported_data_dir}/annot_df.tsv'
                history_df_fn = f'{load_existing_exported_data_dir}/history_df.tsv'

                annot_headers = pd.read_csv(annot_df_fn, sep='\t', nrows=0, index_col=0)
                annot_df = pd.read_csv(annot_df_fn, sep='\t', index_col=0,
                                       converters={h: parse_lists for h in annot_headers})
                history_headers = pd.read_csv(history_df_fn, sep='\t', nrows=0, index_col=0)
                history_df = pd.read_csv(history_df_fn, sep='\t', index_col=0,
                                       converters={h: parse_lists for h in history_headers})

            if isinstance(history_df, pd.DataFrame):
                assert 'index' in history_df.columns

            if isinstance(annot_df, pd.DataFrame) and (annot_df.index.dtype == np.int64):
                warnings.warn(
                    f'annot_df has integers in the index. ' + 
                    f'Check the table is formatted correctly such that the subjects (ie sample, patient ids) are in the first column.\n' + \
                    f'annot_df.index.dtype == np.int64' 
                )
                
            data = self.gen_data(
                description=description,
                annot_df=annot_df,
                annot_col_config_dict=annot_col_config_dict,
                history_df=history_df,
                **kwargs
            )

        self.review_data_interface = ReviewDataInterface(
            data_pkl_fn=data_pkl_fn,
            data=data,
            mh=mh
        )

    def set_default_review_data_annotations_configuration(self):
        """
        Sets preconfigured annotation columns and display settings

        Notes
        -----
        Keep default annotation columns but change display options

            ```
            my_reviewer.set_default_review_data_annotations()

            # custom settings for each
            my_reviewer.add_review_data_annotations_app_display(...)
            my_reviewer.add_review_data_annotations_app_display(...)
            ...
            ```

            or overwrite the defaults

            ```
            my_reviewer.set_default_review_data_annotations_configuration()

            # custom settings for each
            my_reviewer.add_review_data_annotations_app_display(...)
            my_reviewer.add_review_data_annotations_app_display(...)
            ...
            ```

        Add or update annotation configurations
            ```
            my_reviewer.set_default_review_data_annotations()
            my_reviewer.add_review_data_annotations(...)
            my_reviewer.add_review_data_annotations(...)
            ```
            or, completely customize
            ```
            # skip my_reviewer.set_default_review_data_annotations()
            my_reviewer.add_review_data_annotations(...)
            my_reviewer.add_review_data_annotations(...)
            ```

        """
        self.set_default_review_data_annotations()
        self.set_default_review_data_annotations_app_display()

    def add_review_data_annotation(self, annot_name: str, review_data_annotation: DataAnnotation):
        """
        See ReviewData.add_annotation

        """
        self.review_data_interface.add_annotation(annot_name, review_data_annotation)
    
    def set_review_app(self, *args, **kwargs):
        """
        Sets the reviewer's ReviewDataApp.

        Parameters
        ----------
        *args:
            See self.gen_review_app
        **kwargs:
            See self.gen_review_app

        """
        self.app = self.gen_review_app(*args, **kwargs)
        
    def add_annotation_display_component(
        self,
        annot_name,
        annot_display_component: AnnotationDisplayComponent
    ):
        
        if annot_name not in self.review_data_interface.data.annot_col_config_dict.keys():
            raise ValueError(f"Invalid annotation name '{annot_name}'. "
                             f"Does not exist in review data object annotation table"
                             "Make sure to add '{annot_name}' as an annotation with self.add_review_data_annotation()"
                            )
            
        data_annot = self.review_data_interface.data.annot_col_config_dict[annot_name]
            
        if (
            (data_annot.annot_value_type not in annot_display_component.default_compatible_types) and 
            (annot_display_component.display_output_format is None)
        ):
            raise ValueError(
                f'{annot_display_component} only takes types {annot_display_component.default_compatible_types}. '
                f'data annotation {annot_name} is type {data_annot.annot_value_type}. '
                'Change app display type or set a display_output_format function.'
            )
            
        if annot_display_component.default_display_value is not None:
            validate_annot_data(data_annot, annot_display_component.default_display_value)
        
        self.annot_app_display_types_dict[annot_name] = annot_display_component
        
    def remove_annotation_display_component(
        self,
        annot_name
    ):
        if annot_name not in self.annot_app_display_types_dict.keys():
            raise ValueError(f"Annote name '{annot_name}' does not already have a set annotation display component. Nothing to remove.")
        
        removed_annot_display_component = self.annot_app_display_types_dict.pop(annot_name)
        return removed_annot_display_component
    
    def add_review_data_annotations_app_display(self,
                                                annot_name: str,
                                                app_display_type: str):
        """
        Set display type for annotations

        Parameters
        ----------
        annot_name: str
            Name of an annotation column in self.review_data.data annotation table.
            The annotation must be configured in self.review_data.data.annot_col_config_dict
        app_display_type: str
            Type of input display for the annotation. See ReviewDataApp.valid_annotation_app_display_types

        """
        
        warnings.warn(f'add_review_data_annotations_app_display() is deprecated. Please use add_annotation_display_component()')
        
        if annot_name not in self.review_data_interface.data.annot_col_config_dict.keys():
            raise ValueError(f"Invalid annotation name '{annot_name}'. "
                             f"Does not exist in review data object annotation table"
                             "Make sure to add '{annot_name}' as an annotation with self.add_review_data_annotation()"
                            )

        if app_display_type not in valid_annotation_app_display_types:
            raise ValueError(f"Invalid app display type {app_display_type}. "
                             f"Valid options are {valid_annotation_app_display_types}")

        # TODO: check if display type matches annotation type (list vs single value)
        
        data_annot = self.review_data_interface.data.annot_col_config_dict[annot_name]
        
        if (data_annot.annot_value_type == 'float') & (app_display_type != 'number'):
            raise ValueError(
                f'Annotation named "{annot_name}" is a float type. '
                'The only compatible app_display_type is "number". '
                f'Input app_display_type was "{app_display_type}".'
            )
            
        if (data_annot.annot_value_type == 'string') & (app_display_type == 'number'):
            raise ValueError(
                f'Annotation named "{annot_name}" is a string type. '
                f'This is not compatible with an input app_display_type of "{app_display_type}".'
            )

        conversion = {
            'text': TextAnnotationDisplay(),
            'textarea': TextAreaAnnotationDisplay(),
            'number': NumberAnnotationDisplay(),
            'checklist': ChecklistAnnotationDisplay(),
            'radioitem': RadioitemAnnotationDisplay(),
            'select': SelectAnnotationDisplay()
        }

        # self.annot_app_display_types_dict[annot_name] = conversion[app_display_type]
        self.add_annotation_display_component(annot_name, conversion[app_display_type])
        
    def add_autofill(self, autofill_button_name: str, fill_value: Union[State, str, float], annot_name: str):
        """
        Configures the option to use the state of a component in the ReviewDataApp to
        autofill the annotation input form.

        There will be one button corresponding to each component if its name is used in any call to self.add_autofill().
        If the user clicks that button, the app will fill the annotation inputs with the specified values or source
        of data.

        Parameters
        ----------
        autofill_button_name: str
            Name of the button for autofilling.
        fill_value: Union[State, str, float]
            The value to fill the corresponding annotation input specified in annot_col.
            - State(id, attribute): It will fill with whatever value is currently in the layout and its
                corresponding attribute ('children', 'value')
            - str, float: Just a default value to fill if the data in the current component is used to autofill.
                Ex. Component A computes results with one method, while Component B computes results with another.
                You want to annotate with method you used to produce the results that fill the annotation inputs.
        annot_name: str
            The name of an annotation in self.review_data.data.annot_col_config_dict

        """
        if autofill_button_name not in self.autofill_dict.keys():
            self.autofill_dict[autofill_button_name] = {annot_name: fill_value}
        else:
            self.autofill_dict[autofill_button_name][annot_name] = fill_value

        # verify 
        self.app.gen_autofill_buttons_and_states(self.review_data_interface, self.autofill_dict)

    def run(self,
            review_data_table_df: pd.DataFrame = None,
            review_data_table_page_size: int = 5,
            collapsable=True,
            mode='external', 
            host='0.0.0.0', 
            port=8050,
            **kwargs
           ):
        """
        Runs the app

        Parameters
        ----------
        review_data_table_df: dataframe with index that matches the index of the reviewer data object's index
        review_data_table_page_size: number of subjects to view
        mode: {'inline', 'external'}, default='external'
        host: str, default='0.0.0.0'
            Host address
        port: int, default=8050
            Port number

        """
        self.app.run(review_data=self.review_data_interface,
                     autofill_dict=self.autofill_dict,
                     annot_app_display_types_dict=self.annot_app_display_types_dict,
                     review_data_table_df=review_data_table_df,
                     review_data_table_page_size=review_data_table_page_size,
                     collapsable=collapsable,
                     mode=mode,
                     host=host,
                     port=port,
                     **kwargs
                    )

    def get_data_attribute(self, attribute: str):
        return getattr(self.review_data_interface.data, attribute)

    def list_data_attributes(self):
        return vars(self.review_data_interface.data).keys()

    def get_annot(self):
        if not self.review_data_interface.mh.metadata['freeze_data']:
            warnings.warn(
                'Data is not frozen. Annotations will not be saved. '
                'Please freeze data in the dashboard to save annotations.'
            )
        return self.get_data_attribute('annot_df')

    def get_history(self):
        if not self.review_data_interface.mh.metadata['freeze_data']:
            warnings.warn(
                'Data is not frozen. History will not be saved. '
                'Please freeze data in the dashboard to save history.'
            )
        return self.get_data_attribute('history_df')
    
    def export_data(self, path: Union[str, Path], export_by_day=False, dry_run=True, **kwargs):
        """
        Export annotation and history tables to tsv files
        
        Parameters
        ----------
        path: str, Path
            Path to directory to export files. ie /path/to/export_files
        export_by_day: bool
            Optionally add a date to the end of the directory. 
            Currently set to using the day so each day a different directory will be used
        dry_run: bool
        """
        export_dir = f'{path}/{date.today()}' if export_by_day else path

        is_gsurl = path[:5] == 'gs://'

        if not is_gsurl:
            if not os.path.isdir(export_dir):
                print(f'Making new directory {export_dir}')
                os.mkdir(export_dir)
            else:
                warnings.warn(f'Directory {export_dir} already exists')
            
        if not dry_run:
            self.review_data_interface.export_data(export_dir, **kwargs)
            print(f"Exported to {export_dir}")
        else:
            print(f"Export directory will be {export_dir}. Nothing exported yet.")

    def clear_cache(self):
        # All objects collected
        objects = [i for i in gc.get_objects() if isinstance(i, functools._lru_cache_wrapper)]
          
        # All objects cleared
        for object in objects:
            object.cache_clear()


def parse_lists(x):
    """Parses the annotation item, returning a list of items split by commas, or itself.

    Parameters
    ----------
    x: (str, double)
        annotation item to be parsed

    Returns
    -------
    (str, List, double): annotation item, split into a list if needed
    """
    if not x:  # NaN
        return np.NaN  # does AnnoMate expect NaN or empty string??
    elif (x[0] != '[') and (x[-1] != ']'):  # anything not formatted as a list with two brackets
        return x
    elif x == '[]':  # empty list
        return list()
    else:
        return x.strip("[]").replace("'", "").split(", ")



"""Example Reviewer Description
A basic reviewer for the AnnoMate tutorial.
Uses simulated data from simulated_data directory
"""
from AnnoMate.Data import DataAnnotation
from AnnoMate.DataTypes.GenericData import GenericData
from AnnoMate.ReviewDataApp import ReviewDataApp, AppComponent
from AnnoMate.AnnotationDisplayComponent import *
from AnnoMate.ReviewerTemplate import ReviewerTemplate

import pandas as pd
from dash import dcc
from dash import html
from dash.dependencies import Input, Output, State
from typing import Dict, List
import plotly.express as px
import dash_bootstrap_components as dbc


# For pickling to work, need to explicitly define function
def validate_purity(x):
    return x < 0.5


class ExampleReviewer(ReviewerTemplate):
    def gen_data(self,
                 description: str,
                 annot_df: pd.DataFrame,
                 annot_col_config_dict: Dict,
                 history_df: pd.DataFrame,
                 sample_df: pd.DataFrame,
                 preprocessing_str: str,
                 index: List = None,
                 ) -> GenericData:
        """
        Creates a dataframe with index corresponding to the primary dataframe's index

        Parameters
        ----------
        sample_df: pd.DataFrame
            dataframe with samples data and sample ids in the index
        preprocessing_str: str
            Example of preprocessing data
        """
        sample_df['new_column'] = preprocessing_str

        if index is None:
            index = sample_df.index.tolist()

        return GenericData(index=index,
                           description=description,
                           df=sample_df,
                           annot_df=annot_df,
                           annot_col_config_dict=annot_col_config_dict,
                           history_df=history_df
                           )

    def set_default_review_data_annotations(self):
        self.add_review_data_annotation('Notes', DataAnnotation('string'))
        self.add_review_data_annotation('Flag', DataAnnotation('string', options=['Keep', 'Remove'], default='Keep'))

    def gen_review_app(self, mut_file_col: str, sample_cols: List[str]) -> ReviewDataApp:
        """
        Example ReviewDataApp

        Parameters
        ----------
        mut_file_col: str
            column in data object's df dataframe that contains file paths to mutation file for current sample
        sample_cols: List[str]
            List of columns in data object's df dataframe to display per sample being reviewed
        """
        app = ReviewDataApp()

        def gen_data_summary_table(data: GenericData, idx, cols):
            df = data.df
            r = df.loc[idx]
            return [[html.H1(f'{r.name} Data Summary'), dbc.Table.from_dataframe(r[cols].to_frame().reset_index())]]

        app.add_component(AppComponent(name='sample-info-component',
                                       layout=html.Div(children=[html.H1('Data Summary'),
                                                                 dbc.Table.from_dataframe(df=pd.DataFrame())],
                                                       id='sample-info-component'
                                                       ),
                                       callback_output=[Output('sample-info-component', 'children')],
                                       new_data_callback=gen_data_summary_table,
                                       ),
                          cols=sample_cols
                          )

        app.add_table_from_path(data_table_source='df',
                                table_title='Maf file',
                                component_id='maf-component-id',
                                table_fn_col=mut_file_col,
                                table_cols=['sample_id', 'gene', 'vaf', 'cov',
                                            't_alt_count', 't_ref_count'])

        def plot_muts(data: GenericData, idx, color, mut_file_col):
            mut_fn = data.df.loc[idx, mut_file_col]
            maf_df = pd.read_csv(mut_fn, sep='\t')
            fig = px.histogram(maf_df, x='vaf', color_discrete_sequence=[color])
            fig.update_traces()
            return [fig]

        app.add_component(AppComponent(name='Mut vafs',
                                       layout=html.Div([
                                           dcc.Graph(figure={}, id='mut-figure'),
                                           dbc.RadioItems(options=[{'label': c, 'value': c} for c in ['red',
                                                                                                      'blue',
                                                                                                      'green']],
                                                          id='mut-figure-color-radioitem',
                                                          value='red'
                                                          )]),
                                       callback_output=[Output('mut-figure', 'figure')],
                                       callback_input=[Input('mut-figure-color-radioitem', 'value')],
                                       new_data_callback=plot_muts,
                                       internal_callback=plot_muts,
                                       ),
                          mut_file_col=mut_file_col
                          )

        return app

    def set_default_review_data_annotations_app_display(self):
        self.add_annotation_display_component('Notes', TextAreaAnnotationDisplay())
        self.add_annotation_display_component('Flag', RadioitemAnnotationDisplay())

    def set_default_autofill(self):
        self.add_autofill(
            autofill_button_name='example-autofill-button', 
            fill_value='Nothing to say', 
            annot_name='Notes'
        )



"""CNVPlotComponent.py module

Interactive CNV Plot with mutation multiplicity scatterplot

Mutation scatter interactive with mutation table

"""

import pandas as pd
import numpy as np
from dash import dcc, html
from dash.dependencies import Input, Output, State
import dash_bootstrap_components as dbc
import plotly.graph_objects as go
import plotly.subplots
from matplotlib import patches
from scipy.stats import beta
import pickle
import functools

from AnnoMate.ReviewDataApp import AppComponent
from AnnoMate.AppComponents.utils import cluster_color, get_unique_identifier, freezeargs, cached_read_csv
from AnnoMate.DataTypes.PatientSampleData import PatientSampleData

from cnv_suite.visualize import plot_acr_subplots, update_cnv_color_absolute, \
    update_cnv_scatter_sigma_toggle, plot_acr_interactive, add_background
from cnv_suite.utils import calc_cn_levels, apply_segment_data_to_df, get_segment_interval_trees, switch_contigs



def gen_cnv_plot_app_component():
    """Generate CNV Plot app component"""
    return AppComponent(
        'CNV Plot',
        layout=gen_cnv_plot_layout(),
        callback_input=[
            Input('sample-selection-checklist', 'value'),
            Input('sigma_checklist', 'value'),
            Input('cnv-color-radioitem', 'value'),
            Input('absolute-cnv-box', 'value'),  # todo implement as switch, not checkbox
            Input('cnv-button', 'n_clicks'),
            Input('cnv_plot', 'figure'),
            Input('sample-selection-checklist', 'options')
        ],
        callback_output=[
            Output('cnv_plot', 'figure'),
            Output('sample-selection-checklist', 'options'),
            Output('sample-selection-checklist', 'value'),
            Output('cnv-button', 'n_clicks')
        ],
        callback_state_external=[
            State('mutation-selected-ids', 'value'),  # selected rows regardless of filtering
            State('mutation-filtered-ids', 'value')  # all rows in table after filtering
        ],
        new_data_callback=gen_absolute_components,
        internal_callback=internal_gen_absolute_components
    )

def gen_cnv_plot_layout():
    """Generate CNV Plot Component Layout"""
    return html.Div([
        dbc.Row([
            dbc.Col([
                dcc.Graph(
                    id='cnv_plot',
                    figure=go.Figure()
                ),
            ],
            width=10),
            dbc.Col([
                html.H3('Customize Plot'),
                html.H5('Samples:'),
                dcc.Checklist(
                    id='sample-selection-checklist',
                    options=[],
                    value=[],
                    labelStyle={'display': 'block'}
                ),
                html.P(''),
                html.H5('Sigmas:'),
                dcc.Checklist(
                    id='sigma_checklist',
                    options=['Show CNV Sigmas'],
                    value=['Show CNV Sigmas']
                ),
                html.P(''),
                html.H5('Colors:'),
                dcc.RadioItems(
                    id='cnv-color-radioitem',
                    options=['Blue/Red', 'Difference', 'Cluster', 'Clonal/Subclonal', 'Black'],
                    value='Difference',
                    labelStyle={'display': 'block'}
                ),
                html.P(''),
                html.H5('Scale:'),
                dcc.Checklist(
                    options=['Display Absolute CN'],
                    value=['Display Absolute CN'],
                    id='absolute-cnv-box'
                ),
                html.P(''),
                html.Button('Reload Copy Number Plots', id='cnv-button')
            ], width=2)
        ]),
    ])

csize = {'1': 249250621, '2': 243199373, '3': 198022430, '4': 191154276, '5': 180915260,
        '6': 171115067, '7': 159138663, '8': 146364022, '9': 141213431, '10': 135534747,
        '11': 135006516, '12': 133851895, '13': 115169878, '14': 107349540, '15': 102531392,
        '16': 90354753, '17': 81195210, '18': 78077248, '19': 59128983, '20': 63025520,
        '21': 48129895, '22': 51304566, '23': 156040895, '24': 57227415}

def calculate_error(alt, ref, purity, percentile):
    """Calculate error for mutation scatter error bars"""
    if alt == 0:
        return 0
    else:
        return (beta.ppf(percentile, alt, ref) - alt / (alt + ref)) / purity

def gen_mut_scatter(maf_df, mut_sigma, sample, maf_cluster_col, maf_hugo_col, maf_chromosome_col, maf_start_pos_col, maf_variant_class_col, maf_protein_change_col):
    """Generate mutation scatterplot trace.

    Parameters
    ----------
    maf_df
        DataFrame from maf_fn filtered by the mutation table filtering dropdowns
    mut_sigma : bool
        sigmas value based on the sigma checkbox
    sample
        sample selected in the sample selection checkbox

    Returns
    -------
    mut_scatter : go.Scatter

    """
    # cluster_assignment = maf_df.columns[maf_df.columns.isin(['Cluster_Assignment', 'cluster'])][0]

    mut_scatter = go.Scatter(
        x=maf_df['x_loc'],
        y=maf_df['multiplicity_ccf'],
        mode='markers',
        marker_size=5,
        marker_color=maf_df['cluster_color'] if maf_cluster_col in list(maf_df) else 'Black',
        name=f'Mutations ({sample})',
        error_y=dict(
            type='data',
            array=maf_df['error_top'],
            arrayminus=maf_df['error_bottom'],
            color='gray',
            visible=mut_sigma,
            width=0
        ),
        customdata=np.stack((
            maf_df[maf_hugo_col].tolist(),
            maf_df[maf_chromosome_col].tolist(),
            maf_df[maf_start_pos_col].tolist(),
            maf_df['VAF'].tolist() if 'VAF' in maf_df else ['N/A']*len(maf_df),
            maf_df[maf_cluster_col].tolist() if maf_cluster_col in maf_df else ['N/A']*len(maf_df),
            maf_df['Variant_Type'].tolist() if 'Variant_Type' in maf_df else ['N/A']*len(maf_df),
            maf_df[maf_variant_class_col].tolist() if 'Variant_Classification' in maf_df else ['N/A']*len(maf_df),
            maf_df[maf_protein_change_col] if 'Protein_change' in maf_df else ['N/A']*len(maf_df)),
            axis=-1
        ),
        hovertemplate='<extra></extra>' +
                     'Gene: %{customdata[0]} %{customdata[1]}:%{customdata[2]} <br>' +
                     'Variant: %{customdata[5]}, %{customdata[6]} <br>' +
                     'Protein Change: %{customdata[7]} <br>' +
                     'Multiplicity: %{y:.3f} <br>' +
                     'VAF: %{customdata[3]:.3f} <br>' +
                     'Cluster: %{customdata[4]:d}',
        showlegend=False
    )

    return mut_scatter

@freezeargs
@functools.lru_cache(maxsize=32)
def gen_seg_figure(cnv_seg_fn, csize, purity=None, ploidy=None):
    """Generate a CNV Plot from given seg file, purity, and ploidy

    Parameters
    ----------
    cnv_seg_fn: str
        Filename for sample seg file
    csize: dict
        Dictionary with contig sizes
    purity: float
        Tumor purity for this sample (optional)
    ploidy: float
        Tumor ploidy for this sample (optional)

    Returns
    -------
    (cnv_plot, cnv_seg_df_mod, start_trace, end_trace): (plotly.Figure, pd.DataFrame, int, int)
    """
    cnv_seg_df = cached_read_csv(cnv_seg_fn, sep='\t') 
    cnv_plot, cnv_seg_df_mod, start_trace, end_trace = plot_acr_interactive(cnv_seg_df, csize,
                                                                            purity=purity, ploidy=ploidy)

    return cnv_plot, cnv_seg_df_mod, start_trace, end_trace


@freezeargs
@functools.lru_cache(maxsize=16)
def gen_participant_cnv_and_maf(
    cnv_seg_filenames, 
    maf_fn, 
    sample_names, 
    csize, 
    purity_dict, 
    ploidy_dict, 
    maf_start_pos_col, 
    maf_sample_id_col, 
    maf_chromosome_col, 
    maf_cluster_col
):
    """Generate a CNV Plot to be stored in a pickle file

    Parameters
    ----------
    cnv_seg_filenames: list
        List of filenames for the sample cnv seg files
    maf_fn: str
        Filename for the participant maf file
    sample_names: list
        List of sample names for this participant
    csize: dict
        Dictionary with contig sizes
    purity_dict: dict
        Dictionary with purity values for this participant, keys given by sample_id
    ploidy_dict: dict
        Dictionary with ploidy values for this participant, keys given by sample_id

    Returns
    -------
    (participant_maf_df, cnv_plot_dict, cnv_seg_dict, trace_dict): (pd.DataFrame, dict, dict, dict)
    """
    cnv_seg_dict = {}
    cnv_plot_dict = {}
    trace_dict = {}
    for cnv_seg_fn, sample in zip(cnv_seg_filenames, sample_names):
        purity = purity_dict[sample]
        ploidy = ploidy_dict[sample]
        cnv_plot, cnv_seg_df_mod, start_trace, end_trace = gen_seg_figure(cnv_seg_fn, csize, purity=purity, ploidy=ploidy)
        cnv_seg_df_mod['Sample_ID'] = sample
        cnv_seg_dict[sample] = cnv_seg_df_mod
        cnv_plot_dict[sample] = cnv_plot
        trace_dict[sample] = (start_trace, end_trace)

    seg_trees = get_segment_interval_trees(pd.concat(cnv_seg_dict.values()))
    participant_maf_df = gen_maf(maf_fn, purity_dict, ploidy_dict, seg_trees, maf_start_pos_col, maf_sample_id_col, maf_chromosome_col, maf_cluster_col)

    return participant_maf_df, cnv_plot_dict, cnv_seg_dict, trace_dict


def gen_maf(maf_fn, purity_dict, ploidy_dict, seg_trees, maf_start_pos_col, maf_sample_id_col, maf_chromosome_col, maf_cluster_col):
    """

    Parameters
    ----------
    maf_fn: str
        Filename for the participant maf file
    purity_dict: dict
        Dictionary with purity values for this participant, keys given by sample_id
    ploidy_dict: dict
        Dictionary with ploidy values for this participant, keys given by sample_id
    seg_trees: list of IntervalTrees
        IntervalTree for each contig with copy number data

    Returns
    -------
    maf_df: pd.DataFrame
        Modified maf for this participant, including copy number data and additional annotations
    """
    maf_df = cached_read_csv(maf_fn, sep='\t')
    # start_pos = maf_df.columns[maf_df.columns.isin(['Start_position', 'Start_Position'])][0]
    alt = maf_df.columns[maf_df.columns.isin(['Tumor_Seq_Allele2', 'Tumor_Seq_Allele'])][0]
    # sample_id_col = maf_df.columns[maf_df.columns.isin(['Tumor_Sample_Barcode', 'Sample_ID', 'sample_id', 'Sample_id'])][0]
    maf_df['id'] = maf_df.apply(lambda x: get_unique_identifier(x, start_pos=maf_start_pos_col, alt=alt), axis=1)

    maf_df['Sample_ID'] = maf_df[maf_sample_id_col]

    maf_sample_names = set(maf_df['Sample_ID'])
    given_sample_names = set(purity_dict.keys())
    if len(maf_sample_names & given_sample_names) == 0:
        if len(set([p[:-5] for p in maf_sample_names]) & given_sample_names) == 0:
            raise ValueError("Maf sample names don't match what is given")
        else:
            # remove '_pair' from sample ids in maf
            maf_df['Sample_ID'] = maf_df['Sample_ID'].apply(lambda x: x[:-5])

    # maf_df = switch_contigs(maf_df)
    # .replace giving: Series.replace cannot use dict-like to_replace and non-None value ?? todo
    maf_df[maf_chromosome_col] = maf_df[maf_chromosome_col].apply(lambda x: '23' if x == 'X' else '24' if x == 'Y' else x)

    maf_df = apply_segment_data_to_df(maf_df, seg_trees)
    maf_df.set_index('id', inplace=True, drop=False)

    c_values = [calc_cn_levels(pur, plo) for pur, plo in zip(purity_dict.values(), ploidy_dict.values())]
    c_0 = {sample: c[0] for sample, c in zip(purity_dict.keys(), c_values)}
    c_delta = {sample: c[1] for sample, c in zip(purity_dict.keys(), c_values)}

    maf_df['purity'] = maf_df['Sample_ID'].replace(purity_dict)
    maf_df['ploidy'] = maf_df['Sample_ID'].replace(ploidy_dict)
    maf_df['c_0'] = maf_df['Sample_ID'].replace(c_0)
    maf_df['c_delta'] = maf_df['Sample_ID'].replace(c_delta)

    # cluster_assignment = maf_df.columns[maf_df.columns.isin(['Cluster_Assignment', 'cluster'])][0]
    maf_df[maf_chromosome_col] = maf_df[maf_chromosome_col].astype(int)
    if maf_cluster_col in maf_df:
        maf_df['cluster_color'] = maf_df[maf_cluster_col].astype(int).apply(lambda x: cluster_color(x))

    c_size_cumsum = np.cumsum([0] + list(csize.values()))
    maf_df['x_loc'] = maf_df.apply(lambda x: c_size_cumsum[x[maf_chromosome_col] - 1] + x[maf_start_pos_col], axis=1)
    maf_df['VAF'] = maf_df['t_alt_count'] / (maf_df['t_alt_count'] + maf_df['t_ref_count'])

    maf_df['mu_major_adj'] = (maf_df['mu_major'] - maf_df['c_0']) / maf_df['c_delta']
    maf_df['mu_minor_adj'] = (maf_df['mu_minor'] - maf_df['c_0']) / maf_df['c_delta']
    maf_df['multiplicity_ccf'] = maf_df.apply(
        lambda x: x.VAF * (x['purity'] * (x.mu_major_adj + x.mu_minor_adj) + 2 * (1 - x['purity'])) / x['purity'], axis=1)
    # calculate error bars for mutations
    maf_df['error_top'] = maf_df.apply(
        lambda x: calculate_error(x.t_alt_count, x.t_ref_count, x['purity'], 0.975), axis=1)
    maf_df['error_bottom'] = maf_df.apply(
        lambda x: -1 * calculate_error(x.t_alt_count, x.t_ref_count, x['purity'], 0.025), axis=1)

    return maf_df

# Adapting https://github.com/getzlab/cnv_suite/blob/1cfef02623e93d7483841b50dfc598dcfebe577e/cnv_suite/visualize/plot_cnv_profile.py#L93C1-L125C15 to plot background on all subfigures
def updated_plot_acr_subplots(fig_list, title, fig_names, csize, height_per_sample=350, **kwargs):
    """Add each Figure in list to plotly subplots.

    Additional kwargs are passed to plotly's make_subplots method.

    :param height_per_sample: plot height for each sample plot (scales based on size of fig_list); default 350
    :param csize: dict with chromosome sizes, as {contig_name: size}
    :param fig_list: List of plotly.graph_objects.Figure, one for each row
    :param title: Title of plot
    :param fig_names: Title for each subplot (one for each row)
    """
    fig = plotly.subplots.make_subplots(rows=len(fig_list), cols=1,
                                        shared_xaxes=True, subplot_titles=fig_names,
                                        vertical_spacing=0.15/len(fig_list), **kwargs)

    for i, sub_figure in enumerate(fig_list):
        fig.add_traces(sub_figure.data, rows=i + 1, cols=1)
        fig.update_yaxes(fig_list[i].layout.yaxis, row=i+1, col=1)
        # Add chromosome background back in
        add_background(fig, csize.keys(), csize, plotly_row=i+1, plotly_col=1)

    # update height based on subplot number
    fig.update_layout(height=len(fig_list)*height_per_sample + 50,
                      title_text=title,
                      title_font_size=26,
                      plot_bgcolor='white')
    fig.update_xaxes(fig_list[0].layout.xaxis)

    # show x-axis title on only bottom plot
    fig.update_xaxes(title_text='', selector=(lambda x: not x['showticklabels']))

    return fig
    
def gen_cnv_plot(df, idx, sample_selection, sigmas, color, absolute, selected_mutation_rows, filtered_mutation_rows, samples_df, maf_sample_id_col, maf_start_pos_col, maf_chromosome_col, maf_cluster_col, maf_hugo_col, maf_variant_class_col, maf_protein_change_col):
    """Generate CNV Plot with all customizations.

    Parameters
    ----------
    df
        participant level DataFrame
    idx
    sample_selection
        sample selection checkbox value
    sigmas
        sigma checkbox value
    color
        color checkbox value
    absolute
        absolute CN checkbox value
    selected_mutation_rows
        rows selected in the mutation table, None if none selected
    filtered_mutation_rows
        rows filtered in the mutations table, None if none selected
    samples_df
        sample level dataframe
    preprocess_data_dir  TODO: remove
        directory path in which preprocessed CNV pickle files are stored

    Returns
    -------
    cnv_plot : make_subplots()
    sample_list : list of str
        sample checkbox options
    sample_selection_corrected
        first two selections in the sample selection checkbox

    Notes
    -----
    No mutation scatter plot if no purity and ploidy in data
    """
    sample_list = samples_df[samples_df['participant_id'] == idx].sort_values('collection_date_dfd').index.tolist()
    # start with only first sample selected
    sample_selection_corrected = [sample_list[0]] if sample_selection == [] else \
        [s for s in sample_list if s in sample_selection]  # get correct order

    sigmas_val = 'Show CNV Sigmas' in sigmas
    absolute_val = 'Display Absolute CN' in absolute

    # collect Figures and mutation maf
    purity_dict = samples_df.loc[sample_list, 'wxs_purity'].to_dict()
    ploidy_dict = samples_df.loc[sample_list, 'wxs_ploidy'].to_dict()
    cnv_seg_filenames = samples_df.loc[sample_list, 'cnv_seg_fn'].values.tolist()
    maf_fn = df.loc[idx, 'maf_fn']
    participant_maf_df, cnv_plot_dict, cnv_seg_dict, trace_dict = gen_participant_cnv_and_maf(
        cnv_seg_filenames, 
        maf_fn, 
        sample_list, 
        csize, 
        purity_dict, 
        ploidy_dict, 
        maf_start_pos_col, 
        maf_sample_id_col, 
        maf_chromosome_col, 
        maf_cluster_col
    )

    fig_list = []
    for sample_id in sample_selection_corrected:
        cnv_plot = cnv_plot_dict[sample_id]
        this_seg_df = cnv_seg_dict[sample_id]
        update_cnv_color_absolute(cnv_plot, this_seg_df, absolute_val, color,
                                  trace_dict[sample_id][0], trace_dict[sample_id][1])
        if absolute_val:
            cnv_plot.update_yaxes(title_text="Absolute Copy Number")
        else:
            cnv_plot.update_yaxes(title_text="Allelic Copy Ratio")
        fig_list.append(cnv_plot)

    cnv_subplots_fig = updated_plot_acr_subplots(fig_list, 'Copy Number Plots', sample_selection_corrected, csize)
    
    update_cnv_scatter_sigma_toggle(cnv_subplots_fig, sigmas_val)
    
    if selected_mutation_rows:
        participant_maf_df = participant_maf_df.loc[selected_mutation_rows]
    elif filtered_mutation_rows:
        participant_maf_df = participant_maf_df.loc[filtered_mutation_rows]
    # else (if all mutations in table are filtered out and none selected): use all mutations

    for i, sample_id in enumerate(sample_selection_corrected):
        sample_maf_df = participant_maf_df[participant_maf_df[maf_sample_id_col] == sample_id]
        cnv_subplots_fig.add_trace(
            gen_mut_scatter(
                sample_maf_df, 
                sigmas_val, 
                sample_id, 
                maf_cluster_col, 
                maf_hugo_col, 
                maf_chromosome_col, 
                maf_start_pos_col, 
                maf_variant_class_col, 
                maf_protein_change_col
            ), 
            row=i+1, col=1
        )

    return [
        cnv_subplots_fig,
        sample_list,
        sample_selection_corrected
    ]

def gen_absolute_components(
    data: PatientSampleData, 
    idx, sample_selection, sigmas, color, absolute, button_clicks, cnv_plot, sample_list, selected_mutation_rows, filtered_mutation_rows, 
    maf_sample_id_col=None, maf_start_pos_col=None, maf_chromosome_col=None, maf_cluster_col=None, maf_hugo_col=None, maf_variant_class_col=None, 
    maf_protein_change_col=None):
    """Generate CNV plot - new data callback.
    
    Parameters
    ----------
    data
        PatientSampleData containing participant and sample dfs
    idx
        Index - current participant
    sample_selection
        Samples selected in checkboxes to be displayed 
    sigmas
        Sigma checkbox value (checked or unchecked)
    color
        Color seleceted in checkboxes to be displayed 
    absolute
        Display Absolute CN checkbox value (checked or unchecked)
    button_clicks 
        Submit button response 
    cnv_plot
    sample_list 
    selected_mutation_rows 
        Rows selected in the mutation table 
    filtered_mutation_rows 
        Rows displayed in the mutation table after filtering 
    maf_sample_id_col
        kwarg - Name of the sample id column in the maf file 
    maf_start_pos_col
        kwarg - Name of the start position column in the maf file
    maf_chromosome_col
        kwarg - Name of the chromosome column in the maf file
    maf_cluster_col
        kwarg - Name of the cluster assignment column in the maf file  
    maf_hugo_col
        kwarg - Name of the hugo symbol column in the maf file 
    maf_variant_class_col
        kwarg - Name of the variant classification column in the maf file 
    maf_protein_change_col
        kwarg - Name of the protein change column in the maf file 

    Returns
    -------
    Corresponding values to callback outputs (in corresponding order)

    """
    df = data.participant_df
    samples_df = data.sample_df

    # when changing participants, show all mutations at first
    # the filtered and selected mutations input to this function are from the old (previous participant's) MutationTable
    filtered_mutation_rows = None
    selected_mutation_rows = None

    cnv_plot, sample_list, sample_selection = gen_cnv_plot(df, idx, [], sigmas, color, absolute, selected_mutation_rows, filtered_mutation_rows, samples_df, maf_sample_id_col, maf_start_pos_col, maf_chromosome_col, maf_cluster_col, maf_hugo_col, maf_variant_class_col, maf_protein_change_col)
    button_clicks = None

    return [
        cnv_plot,
        sample_list,
        sample_selection,
        button_clicks
    ]

def internal_gen_absolute_components(data: PatientSampleData, idx, sample_selection, sigmas, color, absolute, button_clicks, cnv_plot, sample_list, selected_mutation_rows, filtered_mutation_rows, maf_sample_id_col=None, maf_start_pos_col=None, maf_chromosome_col=None, maf_cluster_col=None, maf_hugo_col=None, maf_variant_class_col=None, maf_protein_change_col=None):
    """Generate CNV plot - internal callback.
    
    Parameters
    ----------
    data
        PatientSampleData containing participant and sample dfs
    idx
        Index - current participant
    sample_selection
        Samples selected in checkboxes to be displayed 
    sigmas
        Sigma checkbox value (checked or unchecked)
    color
        Color seleceted in checkboxes to be displayed 
    absolute
        Display Absolute CN checkbox value (checked or unchecked)
    button_clicks 
        Submit button response 
    cnv_plot
    sample_list 
    selected_mutation_rows 
        Rows selected in the mutation table 
    filtered_mutation_rows 
        Rows displayed in the mutation table after filtering 
    maf_sample_id_col
        kwarg - Name of the sample id column in the maf file 
    maf_start_pos_col
        kwarg - Name of the start position column in the maf file
    maf_chromosome_col
        kwarg - Name of the chromosome column in the maf file
    maf_cluster_col
        kwarg - Name of the cluster assignment column in the maf file  
    maf_hugo_col
        kwarg - Name of the hugo symbol column in the maf file 
    maf_variant_class_col
        kwarg - Name of the variant classification column in the maf file 
    maf_protein_change_col
        kwarg - Name of the protein change column in the maf file 

    Returns
    -------
    Corresponding values to callback outputs (in corresponding order)

    """
    df = data.participant_df
    samples_df = data.sample_df

    if button_clicks != None:
        cnv_plot, sample_list, sample_selection = gen_cnv_plot(df, idx, sample_selection, sigmas, color, absolute, selected_mutation_rows, filtered_mutation_rows, samples_df, maf_sample_id_col, maf_start_pos_col, maf_chromosome_col, maf_cluster_col, maf_hugo_col, maf_variant_class_col, maf_protein_change_col)
        button_clicks = None

    return [
        cnv_plot,
        sample_list,
        sample_selection,
        button_clicks
    ]



from abc import ABC, abstractmethod
from .Data import DataAnnotation, valid_annotation_types
import dash_bootstrap_components as dbc
from dash import dcc
import numpy as np
from typing import List

class AnnotationDisplayComponent(ABC):
    
    def __init__(self, default_display_value=None, display_output_format=None, **kwargs):
        self.display_output_format = display_output_format
        self.default_display_value = default_display_value
        self.default_compatible_types = valid_annotation_types
        self.kwargs = kwargs
        
    @abstractmethod
    def gen_input_component(self, data_annot: DataAnnotation, component_id: str):
        pass
    
    def __str__(self):
        return self.__class__.__name__
    
class MultiValueAnnotationDisplayComponent(AnnotationDisplayComponent):
    
    def __init__(self, default_display_value=None, display_output_format=None, **kwargs):
        super().__init__(default_display_value=default_display_value, display_output_format=display_output_format, **kwargs)
        self.default_compatible_types = ['multi']
        
    @abstractmethod
    def gen_input_component(self, data_annot: DataAnnotation, component_id: str):
        pass
    
    
class TextAnnotationDisplay(AnnotationDisplayComponent):
    
    def gen_input_component(self, data_annot: DataAnnotation, component_id: str):
        return dbc.Input(
            type="text",
            id=component_id,
            value=self.default_display_value,
            **self.kwargs
        )
    
class TextAreaAnnotationDisplay(AnnotationDisplayComponent):
    
    def gen_input_component(self, data_annot: DataAnnotation, component_id: str):
        return dbc.Textarea(
            size="lg", 
            id=component_id,
            value=self.default_display_value,
            **self.kwargs
        )
    
class NumberAnnotationDisplay(AnnotationDisplayComponent):
    def __init__(self, default_display_value=None, display_output_format=None, empty_values: List =None):
        super().__init__(default_display_value=default_display_value, display_output_format=display_output_format)
        self.default_compatible_types = ['float', 'int']

        self.empty_values = ['', None, 'None', 'NaN'] if empty_values is None else empty_values
        if self.display_output_format is None:
            self.display_output_format = lambda x: np.nan if x in self.empty_values else x
    
    def gen_input_component(self, data_annot: DataAnnotation, component_id: str):
        return dbc.Input(
            type="number",
            id=component_id,
            value=self.default_display_value,
            **self.kwargs
        )
    
class ChecklistAnnotationDisplay(MultiValueAnnotationDisplayComponent):
    
    def gen_input_component(self, data_annot: DataAnnotation, component_id: str):
        return dbc.Checklist(
            options=[{"label": f, "value": f} for f in data_annot.options],
            id=component_id, 
            value='' if self.default_display_value is None else self.default_display_value, # Checklist does not take NoneType values
            **self.kwargs
        ) 
    
class RadioitemAnnotationDisplay(AnnotationDisplayComponent):
    
    def __init__(self, default_display_value=None, display_output_format=None, **kwargs):
        super().__init__(default_display_value=default_display_value, display_output_format=display_output_format, **kwargs)
        self.default_compatible_types = ['string']
    
    def gen_input_component(self, data_annot: DataAnnotation, component_id: str):
        return dbc.RadioItems(
            options=[{"label": f, "value": f} for f in data_annot.options],
            value=self.default_display_value,
            id=component_id,
            **self.kwargs
        )
    
class SelectAnnotationDisplay(AnnotationDisplayComponent):
    
    def __init__(self, default_display_value=None, display_output_format=None, **kwargs):
        super().__init__(default_display_value=default_display_value, display_output_format=display_output_format, **kwargs)
        self.default_compatible_types = ['string']
    
    def gen_input_component(self, data_annot: DataAnnotation, component_id: str):
        return dbc.Select(
            options=[{"label": f, "value": f} for f in data_annot.options],
            value=self.default_display_value,
            id=component_id,
            **self.kwargs
        )


class MultiValueSelectAnnotationDisplay(MultiValueAnnotationDisplayComponent):
    
    def gen_input_component(self, data_annot: DataAnnotation, component_id: str):
        
        return dcc.Dropdown(
            options=[{"label": f, "value": f} for f in data_annot.options],
            multi=True,
            id=component_id,
            value=self.default_display_value,
            **self.kwargs
        )
        


"""
DataTableComponents module contains methods to generate components for displaying table information
"""
import pandas as pd
import numpy as np
from dash import dcc, html
from dash.dependencies import Input, Output, State
import dash_bootstrap_components as dbc
import plotly.graph_objects as go
from scipy.stats import beta
import pickle

from AnnoMate.ReviewDataApp import AppComponent
from AnnoMate.AppComponents.utils import cluster_color, get_unique_identifier
from AnnoMate.DataTypes.GenericData import GenericData


def gen_console_links(gsurl: str, link_display_name=None):
    link_display_name = gsurl if link_display_name is None else link_display_name
    
    return html.A(
        html.P(link_display_name),
        href=f"https://console.cloud.google.com/storage/browser/_details/{gsurl.split('/', 2)[-1]}",
        target="_blank"
    )

def gen_annotated_data_info_table(
    data: GenericData,
    data_id,
    cols,
    data_attribute,
    generate_console_links=True,
    console_link_col_name='Console_link',
    link_display_name='Click here',
):
    """
    generate_console_links: Whether you want to convert gsurls to console links
    console_link_col_name: specify the column name where the links will be displayed
    """
    data_df = getattr(data, data_attribute)
    
    if not isinstance(data_df, pd.DataFrame):
        raise ValueError(f'data_attribution {data_attribute} from data object is not a pandas dataframe.')
        
    if data_df.index.tolist() != data.annot_df.index.tolist():
        raise ValueError(f'data.{data_attribute} index does not match the data.annot_df index.')
    
    r = data_df.loc[data_id]
    tmp_data_df = r[cols].to_frame()
    tmp_data_df[r.name] = tmp_data_df[r.name].astype(str)
            
    if generate_console_links:
        
        tmp_data_df[console_link_col_name] = tmp_data_df[r.name].apply(
            lambda url: gen_console_links(url, link_display_name) if 'gs://' in url else ''
        )

    return [[html.H2(f'{r.name} Data Summary'), dbc.Table.from_dataframe(tmp_data_df.reset_index())]]

def gen_annotated_data_info_table_layout(table_id):
    return [
        html.Div(
            children=[
                html.H2('Data Summary'),
                dbc.Table.from_dataframe(df=pd.DataFrame())
            ],
            id=table_id
        )
    ]

def gen_annotated_data_info_table_component(
    component_name='Annotated Data Information table',
    table_id='annot-data-info-component'
):
    """
    Generate a non-interactive table related to the subjects you are reviewing. 
    The table can include hyperlinks to any data that references a gsurl (ie gs://). 
    See gen_annotated_data_info_table() for details.
    """
    return AppComponent(
        component_name,
        layout=gen_annotated_data_info_table_layout(table_id=table_id),
        callback_output=[Output(table_id, 'children')],
        new_data_callback=gen_annotated_data_info_table
    )
    



import pandas as pd
from datetime import datetime
import os
import numpy as np
import warnings
import pickle
from pathlib import Path
from typing import List, Dict, Union
from AnnoMate.Data import Data, DataAnnotation, validate_annot_data
from AnnoMate.MetadataHandler import MetadataHandler


class ReviewDataInterface:
    
    def __init__(self,
                 data_pkl_fn: Union[str, Path],
                 data: Data,
                 mh: MetadataHandler,):
        """
        Object that saves, loads, and edits Data objects

        Parameters
        ----------
        data_pkl_fn: Union[str, Path]
            pickle file to save/load data object from
        data: Data
            data object with the data to review

        Notes
        -----

        If data_pkl_fn already exists, it will only load that file and ignore whatever the parameter data is.
        This is to prevent accidentally overwriting annotations and the data being currently reviewed.
        """
        self.data_pkl_fn = data_pkl_fn
        self.mh = mh
        if os.path.exists(data_pkl_fn) and mh.metadata['freeze_data']:
            f = open(data_pkl_fn, 'rb')
            self.data = pickle.load(f)
            f.close()
            warnings.warn(f"Loading existing data pkl file")
        else:
            self.data = data

        self.save_data()

    def save_data(self):
        """
        Saves Data object to pickle file
        """
        f = open(self.data_pkl_fn, 'wb')
        pickle.dump(self.data, f, 2)
        f.close()
        
    def add_annotation(self,
                       annot_name: str,
                       data_annot: DataAnnotation):
        """
        Adds annotation data to the Data object

        Parameters
        ----------
        annot_name: str
            Name of the annotation that will be a column in the Data object's annot_df dataframe
        data_annot: DataAnnotation
            a DataAnnotation to add to the review data object
        """
        self._add_annotations({annot_name: data_annot})
    
    def _add_annotations(self, annot_col_config_dict: Dict):
        """
        Adds or updates the configuration for annotations to be collected in the Data object

        Parameters
        ----------
        annot_col_config_dict: Dict
            A dictionary where the key is the name of the annotation (column) and the value is a DataAnnotation object
        """

        existing_annot_names = [
            annot_name for annot_name, ann in annot_col_config_dict.items() if annot_name in self.data.annot_df.columns
        ]
        new_annot_names = [
            annot_name for annot_name, ann in annot_col_config_dict.items() if annot_name not in existing_annot_names
        ]

        
        # Check new annotations
        new_data_annot = {annot_name: annot_col_config_dict[annot_name] for annot_name in new_annot_names}
        for annot_name, data_annot in new_data_annot.items():
            if not isinstance(data_annot, DataAnnotation):
                raise ValueError(
                    f'Annotation name {annot_name} has invalid value {data_annot} of type {type(data_annot)}. '
                    f'Value in dictionary must be a DataAnnotation object'
                )
            self.data.annot_col_config_dict[annot_name] = data_annot
            
        # update existing annotations
        for existing_annot_name in existing_annot_names:
            try:
                    
                for idx, r in self.data.annot_df.iterrows():
                    validate_annot_data(annot_col_config_dict[existing_annot_name], r[existing_annot_name])
                    
            except ValueError as e:
                raise ValueError(
                    f'Existing data in annotation table (annot_df) column "{existing_annot_name}" are not compatible with new DataAnnotation configuration. '
                    f'Original message: {e}'
                )
                
            # Replace    
            self.data.annot_col_config_dict[existing_annot_name] = annot_col_config_dict[existing_annot_name]
            
            
        self.data.annot_df[list(new_data_annot.keys())] = np.nan
        self.data.history_df[list(new_data_annot.keys())] = np.nan
        
        # Set types
        for name, data_annot in annot_col_config_dict.items():
            try:
                if data_annot.annot_value_type == 'multi':
                    self.data.annot_df[name] = self.data.annot_df[name].fillna('').astype(object)
                elif data_annot.annot_value_type == 'float':
                    self.data.annot_df[name] = self.data.annot_df[name].astype(float)
                elif data_annot.annot_value_type == 'string':
                    self.data.annot_df[name] = self.data.annot_df[name].fillna('').astype(str)
            except ValueError as e:
                raise ValueError(
                    f'Annotation "{name}" has values that are not compatible with new annot_value_type {data_annot.annot_value_type}. '
                    f'If you mean to change the datatype, consider (1) resetting the reviewer if no annotations were made already, '
                    f'or (2) create a new annotation. '
                    f'Full error: {e}'
                )

        self.save_data()
        
    def _update(self, data_idx, dictionary: Dict):
        """
        Update data annotation table with values in dictionary at index data_idx

        Parameters
        ----------
        data_idx:
            Index in self.data.annot_df
        dictionary: Dict
            A dictionary with keys that exist in self.data.annot_df.columns, and values to put in self.data.annot_df
            at data_idx
        """
        if list(self.data.annot_df.loc[data_idx, list(dictionary.keys())].values) != list(dictionary.values()):
            
            with warnings.catch_warnings():
                
                # Catching warning where the annotation value is "multi" (a list type)
                warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 
                
                self.data.annot_df.loc[data_idx, list(dictionary.keys())] = list(dictionary.values())
                dictionary['timestamp'] = datetime.today()
                dictionary['index'] = data_idx
                dictionary['source_data_fn'] = self.data_pkl_fn
                self.data.history_df = pd.concat([self.data.history_df, pd.Series(dictionary).to_frame().T])
                self.save_data()
        else:
            pass
            
    def export_data(self, path: Union[str, Path], attributes_to_export: List = None, verbose=True):
        """
        Export tables in self.data to tsv files in specified directory

        Parameters
        ----------
        path: Union[str, Path]
            local or gsurl path to directory to save object's dataframe objects
        attributes_to_export: List
            Specify which attributes to export
        """
        attributes_to_export = self.data.__dict__.keys() if attributes_to_export is None else attributes_to_export 

        for attribute_name in attributes_to_export:
            x = getattr(self.data, attribute_name)
            if isinstance(x, pd.DataFrame):
                fn = f'{path}/{attribute_name}.tsv'
                if verbose: print(f'Saving {attribute_name} to {fn}')
                x.to_csv(fn, sep='\t')
            else:
                if verbose: print(f'{attribute_name} is not a dataframe. Not exporting.')



import yaml
import os

class MetadataHandler:
    def __init__(self, fn, overwrite=False):
        self.fn = fn
        self.metadata = self.load_metadata(overwrite=overwrite) # dict of metadata

    def load_metadata(self, overwrite=False):
        if not os.path.exists(self.fn) or overwrite:
            with open(self.fn, 'w'): pass
                
        with open(self.fn, 'r') as file:
            try:
                metadata = yaml.safe_load(file)
                return metadata if metadata is not None else {}
            except yaml.YAMLError as exc:
                print(exc)
                return {}
    
    def set_attribute(self, key, value):
        self.metadata[key] = value
        self.save_metadata()

    def save_metadata(self):
        with open(self.fn, 'w') as file:
            yaml.dump(self.metadata, file)


"""PhylogicNDTComponents.py module

PhylogicNDT CCF Plot and Trees implemented in the PatientReviewer and PhylogicNDTReviewer

PhylogicNDT PMF Plot implemented in the PhylogicNDTReviewer

"""

import pandas as pd
import numpy as np
from dash import dcc
from dash import html, ctx
from dash.dependencies import Input, Output, State
import dash_daq as daq
import dash
import dash_bootstrap_components as dbc
import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots
import dash_cytoscape as cyto
import re
import scipy.stats as ss

from AnnoMate.ReviewDataApp import AppComponent
from AnnoMate.AppComponents.utils import cluster_color, get_unique_identifier
from AnnoMate.DataTypes.PatientSampleData import PatientSampleData


# --------------------- PhylogicNDT CCF Plot and Tree ------------------------

def gen_phylogicNDT_app_component():
    """Generate PhylogicNDT CCF Plot and Tree components"""
    return AppComponent(
        'PhylogicNDT Results',
        layout=gen_phylogicNDT_components_layout(),

        callback_input=[
            Input('time-scale-checklist', 'value'),
            Input('tree-dropdown', 'value'),
            Input('save-tree-button', 'n_clicks')
        ],
        callback_output=[
            Output('ccf-plot', 'figure'),
            Output('tree-dropdown', 'options'),
            Output('tree-dropdown', 'value'),
            Output('phylogicNDT-tree-component', 'children'),
            Output('phylogicNDT-tree', 'generateImage')
        ],
        new_data_callback=gen_phylogicNDT_graphics,
        internal_callback=internal_gen_phylogicNDT_graphics
    )

def gen_phylogicNDT_components_layout():
    """Generate phylogicNDT CCF Plot and Tree layout"""
    return html.Div([
        dcc.Checklist(
            id='time-scale-checklist',
            options=['Time Scaled'],
            value=['Time Scaled'],
        ),
        dbc.Container([
            dbc.Row([
                dbc.Col([
                    html.Div([
                        dcc.Graph(
                            id='ccf-plot',
                            figure=go.Figure()
                        ),
                    ])
                ], width=8, align='center'),
                dbc.Col([
                    html.Div(
                        cyto.Cytoscape(
                            id='phylogicNDT-tree',
                            elements=[],
                            style={'width': '100%', 'height': '450px'},
                        ),
                        id='phylogicNDT-tree-component'
                    ),
                    dcc.Dropdown(
                        id='tree-dropdown',
                        options=[]
                    ),
                    html.Button('save', id='save-tree-button'),
                ], width=4, align='center')
            ])
        ])
    ])

def gen_ccf_plot(df, idx, time_scaled, samples_df, maf_participant_id_col, maf_hugo_col, maf_chromosome_col, maf_start_pos_col, maf_cluster_col):
    """Generate CCF plot including treatment bars.

    Parameters
    ----------
    df
        participant level DataFrame
    idx
    time_scaled
        time scaled checkbox value
    samples_df
        Samples dataframe, containing collection date data as 'collection_date_dfd' data

    Returns
    -------
    ccf_plot : make_subplots()

    """
    # todo add more categories
    treatment_category_colors = {
        'Chemotherapy': 'MidnightBlue',
        'Hormone/Endocrine therapy': 'MistyRose',
        'Precision/Targeted therapy': 'Plum',
        'Immunotherapy': 'Orange'
    }

    cluster_df = pd.read_csv(df.loc[idx, 'cluster_ccfs_fn'], sep='\t', usecols=['Cluster_ID', 'Sample_ID',
                                                                                'postDP_ccf_mean', 'postDP_ccf_CI_low',
                                                                                'postDP_ccf_CI_high'])
    samples_list = cluster_df['Sample_ID'].unique()

    # todo replace this with using sif file - to ensure all collection dates are present and correct
    # pull collection dates from sample table, robust to missing values
    timing_data = {sample: samples_df.loc[sample, 'collection_date_dfd'] if sample in samples_df.index else 0 for sample in samples_list}
    samples_in_order = sorted(timing_data.keys(), key=lambda k: int(timing_data[k]))
    ordered_samples_dict = {s: o for s, o in zip(samples_in_order, np.arange(len(samples_in_order)))}

    # apply dates and sample order to cluster df
    cluster_df.loc[:, 'dfd'] = cluster_df['Sample_ID'].apply(lambda s: int(timing_data[s]))
    cluster_df.loc[:, 'order'] = cluster_df['Sample_ID'].apply(lambda s: ordered_samples_dict[s])

    if 'Time Scaled' in time_scaled:
        scatter_x = 'dfd'
        rect_x = 5
    else:
        scatter_x = 'order'
        rect_x = 6

    if 'treatments_fn' in df:
        treatments_df = pd.read_csv(df.loc[idx, 'treatments_fn'], sep='\t', comment='#')
        treatments_in_frame_df = treatments_df[(treatments_df['stop_date_dfd'] >= int(timing_data[samples_in_order[0]])) &
                                           (treatments_df['start_date_dfd'] <= int(timing_data[samples_in_order[-1]]))]

    # get mutation counts
    mut_ccfs = pd.read_csv(df.loc[idx, 'maf_fn'], sep='\t')
    # participant_id = mut_ccfs.columns[mut_ccfs.columns.isin(['Patient_ID', 'Participant_ID', 'participant_id'])][0]
    # start_pos_id = mut_ccfs.columns[mut_ccfs.columns.isin(['Start_position', 'Start_Position'])][0]
    # cluster = mut_ccfs.columns[mut_ccfs.columns.isin(['Cluster_Assignment', 'cluster'])][0]
    mut_count_dict = mut_ccfs.drop_duplicates([
        maf_participant_id_col,
        maf_hugo_col,
        maf_chromosome_col,
        maf_start_pos_col,
        maf_cluster_col
    ]).groupby(maf_cluster_col).count()[maf_participant_id_col].to_dict()

    cluster_colors = [cluster_color(i) for i in cluster_df['Cluster_ID'].unique()]
    cluster_df['Cluster_ID'] = cluster_df['Cluster_ID'].astype(str)

    ccf_plot = make_subplots(rows=2, cols=1, row_heights=[15,1], shared_xaxes=True)

    for c, color in zip(cluster_df['Cluster_ID'].unique(), cluster_colors):
        this_cluster = cluster_df[cluster_df['Cluster_ID'] == c]
        for i in np.arange(this_cluster.shape[0] - 1):
            x = [this_cluster.iloc[i, rect_x], this_cluster.iloc[i + 1, rect_x], this_cluster.iloc[i + 1, rect_x],
                 this_cluster.iloc[i, rect_x], this_cluster.iloc[i, rect_x]]
            y = [this_cluster.iloc[i, 4], this_cluster.iloc[i + 1, 4], this_cluster.iloc[i + 1, 3],
                 this_cluster.iloc[i, 3], this_cluster.iloc[i, 4]]
            # plot points
            legend = False
            if i == 0:
                legend = True

            ccf_plot.add_trace(
                go.Scatter(
                    x=this_cluster[scatter_x],
                    y=this_cluster['postDP_ccf_mean'],
                    legendgroup=f'group{c}',
                    name=c,
                    marker_color=color,
                    mode='markers',
                    showlegend=legend
                ),
                row=1, col=1
            )
            #confidence interval
            ccf_plot.add_trace(
                go.Scatter(
                    x=x,
                    y=y,
                    legendgroup=f'group{c}',
                    name=f'{c}',
                    fill="toself",
                    fillcolor=color,
                    line_color=color,
                    opacity=0.4,
                    mode='none',
                    showlegend=False
                ),
                row=1, col=1
            )
            # line
            ccf_plot.add_trace(
                go.Scatter(
                    x=[this_cluster.iloc[i, rect_x], this_cluster.iloc[i + 1, rect_x]],
                    y=[this_cluster.iloc[i, 2], this_cluster.iloc[i + 1, 2]],
                    legendgroup=f'group{c}',
                    name=f'{c}',
                    line_width=min(mut_count_dict[int(c)], 15),
                    line_color=color,
                    opacity=0.4,
                    showlegend=False
                ),
                row=1, col=1
            )

    ccf_plot.update_traces(marker_size=15)
    ccf_plot.update_layout(plot_bgcolor='rgba(0,0,0,0)')
    ccf_plot.update_layout(legend={'traceorder': 'reversed'})
    ccf_plot.update_yaxes(title='ccf(x)', dtick=0.1, ticks='outside', showline=True, linecolor='black', range=[-0.03,1.05], showgrid=False)
    ccf_plot.update_xaxes(ticks='outside', showline=True, linecolor='black', showgrid=False)
    if 'Time Scaled' in time_scaled:
        ccf_plot.update_xaxes(title='Time (dfd)')
    else:
        ccf_plot.update_xaxes(title='Samples (timing - dfd)', tickvals=np.arange(len(samples_in_order)),
                         ticktext=[f'{s} ({timing_data[s]})' for s in samples_in_order])
    ccf_plot.data = ccf_plot.data[::-1]  # make the circles appear on top layer

    ccf_plot.add_trace(
        go.Scatter(
            x=this_cluster[scatter_x],
            y=[0,0],
            line_width=20,
            line_color='white',
            fill='toself',
            showlegend=False
        ),
        row=2, col=1
    )

    if 'Time Scaled' in time_scaled and 'treatments_fn' in df:
        for start, stop, drug, drug_combo, category, stop_reason, post_status in zip(treatments_in_frame_df.start_date_dfd,
                                                                                     treatments_in_frame_df.stop_date_dfd,
                                                                                     treatments_in_frame_df.drugs,
                                                                                     treatments_in_frame_df.drug_combination,
                                                                                     treatments_df.categories,
                                                                                     treatments_in_frame_df.stop_reason,
                                                                                     treatments_in_frame_df.post_status):
            drug = drug_combo if pd.isna(drug) else drug

            # todo deal with overlapping treatments
            ccf_plot.add_trace(
                go.Scatter(
                    # todo implement 'order' for x
                    x=[max(start, int(timing_data[samples_in_order[0]])), min(stop, int(timing_data[samples_in_order[-1]]))],
                    y=[0,0],
                    line_width=20,
                    line_color=treatment_category_colors[category] if category in treatment_category_colors.keys() else 'gray',
                    fill='toself',
                    hovertemplate = '<extra></extra>' +
                        f'Treatment Regimen: {drug} <br>' +
                        f'Stop Reason: {stop_reason} <br>' +
                        f'Post Status: {post_status}',
                    showlegend=False
                ),
                row=2, col=1
            )
            ccf_plot.add_vline(
                x=max(start, int(timing_data[samples_in_order[0]])),
                line_width=2,
                line_color='black',
                row=2, col=1
            )
            ccf_plot.add_vline(
                    x=min(stop, int(timing_data[samples_in_order[-1]])),
                    line_width=2,
                    line_color='black',
                    row=2, col=1
            )

    ccf_plot.update_yaxes(row=2, visible=False)
    ccf_plot.update_xaxes(row=1, visible=False, showticklabels=False)

    return ccf_plot

def gen_stylesheet(cluster_list):
    """Format PhylogicNDT tree to have correct cluster colors and labels

    Parameters
    ----------
    cluster_list
        list of clusters in given data

    Returns
    -------
    stylesheet : list of dicts

    """
    stylesheet = [
        {
            'selector': 'node',
            'style': {
                'label': 'data(label)',
                'width': '50%',
                'height': '50%',
                'text-halign':'center',
                'text-valign':'center',
                'color': 'white'
            }
        },
        {
            'selector': 'edge',
            'style': {
                'label': 'data(label)',
                'text-halign':'center',
                'text-valign':'center',
                'color': 'black',
                'text-wrap': 'wrap',
                'font-weight': 'bold'
            }
        }
    ]

    for node in cluster_list:
        stylesheet.append({
            'selector': ('node[label = "%s"]' % node),
            'style': {
                'background-color': cluster_color(int(node))
            }
        })
        stylesheet.append({
            'selector': ('edge[target = "%s"]' % f'cluster_{node}'),
            'style': {
                'line-color': cluster_color(int(node))
            }
        })

    return stylesheet

def gen_driver_edge_labels(drivers, cluster_hugo_list):
    """Add driver mutation label to edges of clusters containing that driver

    Parameters
    ----------
    drivers: pd.DataFrame()
        DataFrame from the drivers kwarg file
    cluster_hugo_list: list of str
        list of hugo symbols associated with a particular cluster

    Returns
    -------
    label: str

    """
    label = ''
    for driver in drivers.drivers:
        if driver in cluster_hugo_list:
            label += ('%s \n' % driver)

    return label

def gen_phylogicNDT_tree(df, idx, tree_num, drivers_fn, maf_start_pos_col, maf_cluster_col, maf_hugo_col, tree_meta_col='n_iter'):
    """Generate PhlogicNDT tree and dropdown to choose from all possible trees.

    Parameters
    ----------
    df
        participant level DataFrame
    idx
    tree_num
        number assigned to the chosen tree that is to be displayed
    drivers_fn
        name of the drivers file passed into to gen_review_app as kwarg
    tree_meta_col
        Which column from the tree file to display along with the tree name in the dropdown. See data in 'build_tree_posterior_fn'

    Returns
    -------
    cyto.Cytoscape
        the tree figure
    possible_trees : list of str
        possible tree options for dropdown

    """
    tree_df = pd.read_csv(df.loc[idx, 'build_tree_posterior_fn'], sep='\t')
    maf_df = pd.read_csv(df.loc[idx, 'maf_fn'], sep='\t')
    maf_df.drop_duplicates(subset=maf_start_pos_col, inplace=True)
    if drivers_fn:
        drivers = pd.read_csv(drivers_fn, header=None, names=['drivers'])

    # cluster_assignment_col = maf_df.columns[maf_df.columns.isin(['Cluster_Assignment', 'cluster'])][0]
    cluster_assignments = maf_df[maf_cluster_col].unique().tolist()
    possible_trees = []
    possible_trees_edges = []
    clusters = {}
    cluster_count = {}
    cluster_list = []

    trees = tree_df.loc[:, 'edges']
    for i, tree in enumerate(trees):
        possible_trees_edges.append(tree.split(','))
        tree_label = f'Tree {i+1}'
        if tree_meta_col:
            tree_label = tree_label + f' ({tree_meta_col}={tree_df[tree_meta_col][i]})'
        possible_trees.append(tree_label)

    for i in range(len(cluster_assignments)):
        clusters[cluster_assignments[i]] = [hugo for clust, hugo in zip(maf_df[maf_cluster_col], maf_df[maf_hugo_col]) if clust == cluster_assignments[i]]

    for clust in clusters:
        cluster_count[clust] = len(clusters[clust])

    edges = possible_trees_edges[tree_num]

    for i in edges:
        new_list = i.split('-')
        for j in new_list:
            if (j !='None') & (j not in cluster_list):
                cluster_list.append(j)
    cluster_list = sorted(cluster_list)

    nodes = [{'data': {'id': 'normal', 'label': 'normal'}, 'position': {'x': 0, 'y': 0}}]

    nodes.extend([
        {
            'data': {'id': f'cluster_{cluster}', 'label': cluster},
            'position': {'x': 50 * int(cluster), 'y': -50 * int(cluster)}
        }
        for cluster in cluster_list
    ])

    edges_list = []
    nodes_copy = nodes.copy()
    for edge in edges:
        nodes_copy = edge.split('-')
        if nodes_copy[0]!='None':
            nodes_copy = list(map(int,edge.split('-')))
            edges_list.append(nodes_copy)

    if drivers_fn:
        edges = [{'data': {'source': 'normal', 'target': 'cluster_1', 'label': f'{cluster_count[1]}\n{gen_driver_edge_labels(drivers, clusters[1])}'}}]
        edges.extend([
            {'data': {'source': f'cluster_{edge[0]}', 'target': f'cluster_{edge[1]}', 'label': f'{cluster_count[edge[1]]}\n{gen_driver_edge_labels(drivers, clusters[edge[1]])}'}}
            for edge in edges_list
        ])
    else:
        edges = [{'data': {'source': 'normal', 'target': 'cluster_1', 'label': str(cluster_count[1])}}]
        edges.extend([
            {'data': {'source': f'cluster_{edge[0]}', 'target': f'cluster_{edge[1]}', 'label': str(cluster_count[edge[1]])}}
            for edge in edges_list
        ])

    elements = nodes + edges

    stylesheet = gen_stylesheet(cluster_list)

    return [
        cyto.Cytoscape(
            id='phylogicNDT-tree',
            style={'width': '100%', 'height': '450px'},
            layout={
                'name': 'breadthfirst',
                'roots': '[id="normal"]'
            },
            elements=elements,
            stylesheet=stylesheet,
            userZoomingEnabled=False
        ),
        possible_trees
    ]

def gen_phylogicNDT_graphics(
        data: PatientSampleData, 
        idx, time_scaled, chosen_tree, save_tree_button, 
        drivers_fn=None, maf_participant_id_col=None, maf_hugo_col=None, maf_chromosome_col=None, maf_start_pos_col=None, maf_cluster_col=None,
        tree_meta_col='n_iter'
    ):
    """Generate PhylogicNDT CCF plot and trees - new data callback.

    Parameters
    ----------
    data
        PatientSampleData containing participant and sample dfs
    idx
        Index - current participant
    time_scaled
        Time scaled checkbox input (checked or not checked)
    chosen_tree
        Tree chosen in the tree dropdown
    save_tree_button
        Response from the 'save' button under the tree dropdown being clicked
    drivers_fn
        kwarg - File path to csv file of drivers
    maf_participant_id_col
        kwarg - Name of the participant id column in the maf file
    maf_hugo_col
        kwarg - Name of the hugo symbol column in the maf file 
    maf_chromosome_col
        kwarg - Name of the chromosome column in the maf file 
    maf_start_pos_col
        kwarg - Name of the start position column in the maf file 
    maf_cluster_col
        kwarg - Name of the cluster assignment column in the maf file  

    Returns
    -------
    Corresponding values to callback outputs (in corresponding order)

    """
    df = data.participant_df
    samples_df = data.sample_df

    if df.loc[idx, 'cluster_ccfs_fn']:
        ccf_plot = gen_ccf_plot(df, idx, time_scaled, samples_df, maf_participant_id_col, maf_hugo_col, maf_chromosome_col, maf_start_pos_col, maf_cluster_col)
        tree, possible_trees = gen_phylogicNDT_tree(df, idx, 0, drivers_fn, maf_start_pos_col, maf_cluster_col, maf_hugo_col, tree_meta_col=tree_meta_col)

        return [ccf_plot, possible_trees, possible_trees[0], tree, dash.no_update]
    else:
        return [go.Figure, [], 0, '', dash.no_update]

def internal_gen_phylogicNDT_graphics(
        data: PatientSampleData, 
        idx, time_scaled, chosen_tree, save_tree_button, 
        drivers_fn=None, maf_participant_id_col=None, maf_hugo_col=None, maf_chromosome_col=None, maf_start_pos_col=None, maf_cluster_col=None,
        tree_meta_col='n_iter'
    ):
    """Generate PhylogicNDT CCF plot and trees - interal callback.
    
    Parameters
    ----------
    data
        PatientSampleData containing participant and sample dfs
    idx
        Index - current participant
    time_scaled
        Time scaled checkbox input (checked or not checked)
    chosen_tree
        Tree chosen in the tree dropdown
    save_tree_button
        Response from the 'save' button under the tree dropdown being clicked
    drivers_fn
        kwarg - File path to csv file of drivers
    maf_participant_id_col
        kwarg - Name of the participant id column in the maf file
    maf_hugo_col
        kwarg - Name of the hugo symbol column in the maf file 
    maf_chromosome_col
        kwarg - Name of the chromosome column in the maf file 
    maf_start_pos_col
        kwarg - Name of the start position column in the maf file 
    maf_cluster_col
        kwarg - Name of the cluster assignment column in the maf file  

    Returns
    -------
    Corresponding values to callback outputs (in corresponding order)

    """
    df = data.participant_df
    samples_df = data.sample_df
    if ctx.triggered:
        if ctx.triggered_id != 'save-tree-button':
            if df.loc[idx, 'cluster_ccfs_fn']:
                tree_num = 0
                for n in chosen_tree.split():
                    if n.isdigit():
                        tree_num = int(n)

                ccf_plot = gen_ccf_plot(df, idx, time_scaled, samples_df, maf_participant_id_col, maf_hugo_col, maf_chromosome_col, maf_start_pos_col, maf_cluster_col)
                tree, possible_trees = gen_phylogicNDT_tree(df, idx, tree_num-1, drivers_fn, maf_start_pos_col, maf_cluster_col, maf_hugo_col, tree_meta_col)

                return [ccf_plot, possible_trees, chosen_tree, tree, dash.no_update]
            else:
                return [go.Figure, [], 0, '', dash.no_update]
        else:
            return [dash.no_update, dash.no_update, dash.no_update, dash.no_update, {'type': 'jpg', 'action': 'download'}]


# -------------------------- PhylogicNDT PMF Plot ----------------------------
def gen_ccf_pmf_component():
    return AppComponent(name='CCF pmf Mutation Plot',
                        layout=html.Div([
                            dbc.Row([
                                dbc.Col([
                                    dcc.Graph(id='pmf-plot',
                                              figure=go.Figure())
                                ], width=10),
                                dbc.Col([html.H3('Customize Plot'),
                                         html.H5('Samples:'),
                                         dcc.Checklist(options=[],
                                                       value=[],
                                                       id='sample-selection'),
                                         html.P(''),
                                         html.H5('Group by Cluster:'),
                                         daq.BooleanSwitch(id='group-clusters', on=True),
                                         html.P(''),
                                         html.Button('Submit', id='pmf-button')
                                ], width=2)
                            ])
                        ]),
                        callback_input=[
                            Input('pmf-button', 'n_clicks')
                        ],
                        callback_output=[
                            Output('pmf-plot', 'figure'),
                            Output('sample-selection', 'options'),
                            Output('sample-selection', 'value')
                        ],
                        callback_state=[
                            State('sample-selection', 'value'),
                            State('group-clusters', 'on')
                        ],
                        callback_state_external=[
                            State('mutation-selected-ids', 'value'),  # selected rows
                            State('mutation-filtered-ids', 'value')  # all rows in table after filtering
                        ],
                        new_data_callback=gen_pmf_component,
                        internal_callback=update_pmf_component
                        )


def ccf_pmf_plot(data_df, idx, sample_selection, group_clusters, selected_mut_ids, filtered_mut_ids, maf_sample_id_col, maf_cluster_col, maf_hugo_col, maf_chromosome_col, maf_start_pos_col):

    """Plots the CCF pmf distribution for the chosen mutation(s).

    Notes
    -----
    - Displays the pmf distribution as a normalized histogram
    - Samples are shown in separate rows
    - Clusters displayed with different colors, with adjacent bars
    - Given maf file in column 'maf_fn' in the df must be mut_ccfs file

    TODO
    ----
    - Add a star (*) above the mode for each mutation
    - Add an indication of mean?

    """
    mut_ccfs_df = pd.read_csv(data_df.loc[idx, 'maf_fn'], sep='\t')
    mut_ccfs_df['unique_mut_id'] = mut_ccfs_df.apply(get_unique_identifier, axis=1)  # must be mut_ccfs file with default columns
    mut_ccfs_df.set_index('unique_mut_id', inplace=True, drop=False)

    # Use only the selected mutations unless no mutations selected, then use filtered list
    if selected_mut_ids:
        mut_ccfs_df = mut_ccfs_df.loc[selected_mut_ids].copy()
    elif filtered_mut_ids:
        mut_ccfs_df = mut_ccfs_df.loc[filtered_mut_ids].copy()
    # else (if all mutations in table are filtered out and none selected): use all mutations

    sample_list = mut_ccfs_df[maf_sample_id_col].unique()  # todo ensure sorted by collection date
    sample_selection = sample_list if not sample_selection else sample_selection
    mut_ccfs_df = mut_ccfs_df[mut_ccfs_df[maf_sample_id_col].isin(sample_selection)].copy()

    ccfs_headers = [re.search('.*[01].[0-9]+', i) for i in mut_ccfs_df.columns]
    ccfs_headers = [x.group() for x in ccfs_headers if x]
    ccfs_header_dict = {i: re.search('[01].[0-9]+', i).group() for i in ccfs_headers}

    stacked_muts = mut_ccfs_df.set_index([maf_sample_id_col, 'unique_mut_id', maf_cluster_col])[
        ccfs_headers].stack().reset_index().rename(columns={'level_3': 'CCF', 0: 'Probability'}).replace(
        ccfs_header_dict)
    if group_clusters:
        stacked_muts[maf_cluster_col] = stacked_muts[maf_cluster_col].astype(str)
        fig = px.histogram(stacked_muts, x='CCF', y='Probability', facet_row=maf_sample_id_col, barmode='group',
                           height=300 * len(sample_selection), color=maf_cluster_col, histfunc='avg',
                           color_discrete_map=cluster_color(),
                           category_orders={maf_cluster_col: np.arange(1, 100)})
    else:
        fig = px.histogram(stacked_muts, x='CCF', y='Probability', facet_row=maf_sample_id_col, barmode='group',
                           height=300 * len(sample_selection), color='unique_mut_id',
                           labels={'unique_mut_id': 'Mutation'})
        mut_label_dict = {x['unique_mut_id']: f"{x[maf_hugo_col]} - {x[maf_chromosome_col]}:{x[maf_start_pos_col]}" for idx, x
                          in mut_ccfs_df.drop_duplicates('unique_mut_id').iterrows()}
        fig.for_each_trace(lambda t: t.update(name=mut_label_dict[t.name]))

    fig.update_layout(xaxis_tickangle=0, xaxis_ticklabelstep=5)
    fig.update_yaxes(matches=None)

    return fig, sample_list


def gen_pmf_component(data: PatientSampleData, idx, button_clicks, sample_selection,
                      group_clusters, selected_mut_ids, filtered_mut_ids,
                      maf_sample_id_col=None, maf_cluster_col=None, maf_hugo_col=None, 
                      maf_chromosome_col=None, maf_start_pos_col=None):
    # when changing participants, show all mutations at first
    # the filtered and selected mutations input to this function are from the old (previous participant's) MutationTable
    filtered_mut_ids = None
    selected_mut_ids = None

    fig, sample_list = ccf_pmf_plot(data.participant_df, idx, None, group_clusters, selected_mut_ids, filtered_mut_ids, maf_sample_id_col, maf_cluster_col, maf_hugo_col, maf_chromosome_col, maf_start_pos_col)

    return [fig, sample_list, sample_list]


def update_pmf_component(data: PatientSampleData, idx, button_clicks, sample_selection,
                         group_clusters, selected_mut_ids, filtered_mut_ids, 
                         maf_sample_id_col=None, maf_cluster_col=None, maf_hugo_col=None, 
                         maf_chromosome_col=None, maf_start_pos_col=None):
    fig, sample_list = ccf_pmf_plot(data.participant_df, idx, sample_selection,
                                    group_clusters, selected_mut_ids, filtered_mut_ids, maf_sample_id_col, maf_cluster_col, maf_hugo_col, maf_chromosome_col, maf_start_pos_col)

    return [fig, sample_list, sample_selection]

# ----------------------------- PhylogicNDT Cluster Metrics -------------------------------

def gen_cluster_metrics_component():
    return AppComponent(name='Mutation Types by Cluster',
                        layout=html.Div([
                                    dcc.Graph(id='metric-plot',
                                              figure=go.Figure())
                        ]),
                        callback_output=[
                            Output('metric-plot', 'figure'),
                        ],
                        new_data_callback=gen_cluster_metric_fig,
                        internal_callback=gen_cluster_metric_fig
                        )


def gen_cluster_metric_fig(data: PatientSampleData, idx, maf_variant_type_col=None, maf_variant_class_col=None, maf_cluster_col=None):
    """Generate a figure showing mutation type comparisons across clusters with indication of differences."""
    data_df = data.participant_df

    mut_ccfs_df = pd.read_csv(data_df.loc[idx, 'maf_fn'], sep='\t')
    mut_ccfs_df['unique_mut_id'] = mut_ccfs_df.apply(get_unique_identifier, axis=1)  # mut_ccfs file has default columns
    mut_ccfs_df.drop_duplicates('unique_mut_id', inplace=True)

    # apply functions to specify coding vs. non-coding; silent vs. nonsyn
    mut_ccfs_df['snv_indel'] = mut_ccfs_df[maf_variant_type_col].apply(lambda x: 'SNV' if x == 'SNP' or 'SNV' else 'INDEL')
    mut_ccfs_df['snv_indel'] = pd.Categorical(mut_ccfs_df['snv_indel'], categories=['SNV', 'INDEL'])
    mut_ccfs_df['class'] = mut_ccfs_df[maf_variant_class_col].apply(classify_mut)
    mut_ccfs_df['class'] = pd.Categorical(mut_ccfs_df['class'], categories=['non-coding', 'synonymous', 'non-synonymous'])

    mut_type_counts = mut_ccfs_df.groupby([maf_cluster_col, 'snv_indel']).size().unstack(fill_value=0)
    mut_classes_counts = mut_ccfs_df.groupby([maf_cluster_col, 'class']).size().unstack(fill_value=0)
    mut_classes_counts['coding'] = mut_classes_counts['synonymous'] + mut_classes_counts['non-synonymous']

    mut_counts_df = mut_classes_counts.join(mut_type_counts)
    mut_counts_df_mod = mut_counts_df.copy()
    mut_counts_df_mod.loc['ALL', :] = mut_counts_df_mod.sum()
    mut_counts_df_mod = mut_counts_df_mod.stack().reset_index().rename(columns={'level_1': 'annotation', 0: 'count'})
    mut_counts_df_mod['type'] = mut_counts_df_mod['annotation'].apply(
        lambda x: 'Non/Coding' if 'coding' in x else ('Non/Syn' if 'syn' in x else 'SNV/INDEL'))
    # turn these into pie charts (one for each mut comparison type)
    num_clusters = len(mut_counts_df_mod[maf_cluster_col].unique())
    fig = make_subplots(
        rows=2, cols=num_clusters, 
        specs=[[{"type": "pie"} for _ in range(0, num_clusters)] for _ in range(0, 2)], 
        subplot_titles=[i for i in mut_counts_df_mod[maf_cluster_col].unique()],
        row_titles=[i for i in mut_counts_df_mod['type'].unique()]
    )
    for i, annotations in enumerate(zip(['non-coding', 'synonymous'], ['coding', 'non-synonymous'])):
        for j, clust in enumerate(mut_counts_df_mod[maf_cluster_col].unique()):
            filterby_annotation_df = mut_counts_df_mod[mut_counts_df_mod['annotation'].isin(annotations)]
            filterby_cluster_df = filterby_annotation_df[filterby_annotation_df[maf_cluster_col] == clust]
            filterby_cluster_df.set_index('annotation', inplace=True)

            fig.add_trace(go.Pie(
                values=filterby_cluster_df['count'],
                domain=dict(x=[i/num_clusters, (i+1)/num_clusters]),
                labels=filterby_cluster_df['count'].index,
                legendgroup=i,
                textinfo='value',
            ), row=(i+1), col=(j+1))
            fig.update_layout(height=500)
            
    fig.for_each_annotation(lambda a: a.update(font_size=18))
    fig.for_each_annotation(lambda a: a.update(text=a.text.replace("Cluster_Assignment=", ""), y=1.05),
                            selector={'xanchor': 'center'})
    fig.for_each_annotation(lambda a: a.update(text=a.text.replace("type=", "")))

    # annotate any cluster that is significantly different from all other clusters
    # using fisher exact test to compare each cluster to all other clusters
    fisher_p_df = pd.DataFrame()
    for col1, col2 in zip(['non-coding', 'synonymous'], ['coding', 'non-synonymous']):
        for idx in mut_counts_df.index:
            odds_ratio, p_val = ss.fisher_exact([[mut_counts_df.loc[idx, col1], mut_counts_df[col1].drop(idx).sum()],
                                                 [mut_counts_df.loc[idx, col2], mut_counts_df[col2].drop(idx).sum()]])
            fisher_p_df.loc[idx, col2] = p_val

    cluster_sig = (fisher_p_df < 0.05).any(axis=1)
    for idx in cluster_sig[cluster_sig].index:
        fig.for_each_annotation(lambda a: a.update(text=f'<b>{idx}*<b>', font_color='red'),
                                selector={'text': str(idx)})
    # todo add indication of which category is significantly different

    return [fig]


def classify_mut(variant_class):
    if variant_class in ['lincRNA', 'RNA', 'IGR', "3'UTR", "5'UTR", 'Intron', "5'Flank", "3'Flank", 'intronic']:
        return 'non-coding'
    elif variant_class in ['Silent', 'syn']:
        return 'synonymous'
    else:
        return 'non-synonymous'



from datetime import datetime
import pandas as pd
import numpy as np
# from jupyter_dash import JupyterDash
from dash import jupyter_dash
from dash import dcc, ctx
from dash import html, Dash
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash
import dash_bootstrap_components as dbc
import inspect
from collections import OrderedDict
from typing import Dict
import copy
import warnings
from typing import Union, List, Tuple
from math import floor
from pathlib import Path
import os

from .ReviewDataInterface import ReviewDataInterface
from .Data import DataAnnotation, validate_annot_data
from .AnnotationDisplayComponent import AnnotationDisplayComponent

valid_annotation_app_display_types = ['text',
                                      'textarea',
                                      'number',
                                      'checklist',
                                      'radioitem',
                                      'select']

class AppComponent:
    
    def __init__(self, 
                 name: str, 
                 layout: Union[List, Tuple, html.Div],
                 callback_output: [Output] = [],
                 callback_input: [Input] = [],
                 callback_state: [State] = [],
                 callback_state_external: [State] = [],
                 new_data_callback=None,
                 internal_callback=None,
                 use_name_as_title=True):
        
        """
        Component in the plotly dashboard app. Each component is made up of a layout and callback functions
        defining interactive behaviors

        Parameters
        ----------
        name: str
            Unique name for the component. Will be used as the title displayed in the dashboard
            (preceeding the layout)
        
        layout: html
            dash html layout object (ex. html.Div([html.H1(children="heading", id='heading'), ...])),
        
        callback_output: List[Output]
            List of Output() objects pointing to objects in the layout
            (ex. [Output('heading', 'children')]) where callback function outputs will be stored
                          
        callback_input: List[Input]
            List of Input() objects pointing to objects in the layout
            (ex. [Input('heading', 'children')]). Changes to these specified Input objects
            will activate the callback function and its current value will be used as an input
                          
        callback_state: List[Input]
            List of State() objects pointing to objects in the layout
            (ex. [State('heading', 'children')]). It's value will be used as a parameter
            if ANY components' callback function is activated.

        callback_state_external: List[State]
            List of State() objects pointing to objects in the layout
            (ex. [State('heading', 'children')]) that do NOT exist in the layout,
            but instead from other AppComponent objects included in the same ReviewDataApp.
            Note that there will be no native check if the component id exists until runtime
            instantiation of the ReviewDataApp.
            It's value will be used as a parameter if the component's callback function
            is activated.
                          
        new_data_callback: func
            a function (defined separately or a lambda) that defines how the component will be updated
            when the ReviewData
            object switches to a new index or row to review. Requirements are:
            - The first two parameters will be the ReviewData.data object and the index of
                the current row to be reviewed.
            - The following parameters will be the inputs defined IN ORDER by the order of the
                Inputs() in callback_input
            - The next parameters will be the states defined IN ORDER by the order of the States()
                in callback_state
            - The next parameters will be the states defined IN ORDER by the order of the States()
                in callback_state_external
            - Any remaining parameters have keywords and are at the end
            - **new_data_callbackinternal_callback must have the same signature
                Example:
                    def myfunc(data: Data, idx, input1, input2, state1, **kwargs):...

                    callback_input = [Input('id-of-object-with-value-for-input1', 'value'),
                                      Input('id-of-object-with-value-for-input2', 'children')]

        internal_callback: func
            a function (defined separately or a lambda) that defines how the component will be updated
            when one of the defined callback_input Inputs are changed.
            Same requirements as new_data_callback

        use_name_as_title: bool
            use the `name` parameter as a title for the component.
        """
        
        all_ids = np.array(get_component_ids(layout))
        check_duplicates(all_ids, 'component')

        # preprocess callback_output to include allow_duplicate=True
        if isinstance(callback_output, list):
            new_callback_output = [
                Output(**{
                    'component_id': output.component_id, 
                    'component_property': output.component_property, 
                    'allow_duplicate': True
                }) for output in callback_output
            ]  
        elif isinstance(callback_output, dict):
            new_callback_output = {
                k: Output(**{
                    'component_id': output.component_id, 
                    'component_property': output.component_property, 
                    'allow_duplicate': True
                }) for k, output in callback_output.items()
            }
        else:
            raise ValueError(f'callback output {callback_output} is neither a list or a dictionary.')

        check_duplicate_objects(new_callback_output, 'callback_output_objects')
        callback_output_ids = get_callback_io_ids(new_callback_output, expected_io_type=Output)
        check_callback_io_id_in_list(callback_output_ids, all_ids, ids_type='output_ids', all_ids_type='component_ids')
        
        check_duplicate_objects(callback_input, 'callback_input_objects')
        callback_input_ids = get_callback_io_ids(callback_input, expected_io_type=Input)
        check_callback_io_id_in_list(callback_input_ids, all_ids, ids_type='input_ids', all_ids_type='component_ids')
        
        check_duplicate_objects(callback_state, 'callback_state_objects')
        callback_state_ids = get_callback_io_ids(callback_state, expected_io_type=State)
        check_callback_io_id_in_list(callback_state_ids, all_ids, ids_type='state_ids', all_ids_type='component_ids')

        if internal_callback is not None and \
                inspect.signature(new_data_callback) != inspect.signature(internal_callback):
            raise ValueError(f'new_data_callback and internal_callback do not have the same signature.\n'
                             f'new_data_callback signature:{inspect.signature(new_data_callback)}\n'
                             f'internal_callback signature:{inspect.signature(internal_callback)}')
            
        self.name = name
        if use_name_as_title:
            self.layout = html.Div([html.H1(name), html.Div(layout)])
        else:
            self.layout = html.Div(layout)
            
        self.all_component_ids = all_ids
        self.callback_output = new_callback_output
        self.callback_input = callback_input
        self.callback_state = callback_state
        self.callback_state_external = callback_state_external
        
        self.new_data_callback = new_data_callback
        self.internal_callback = internal_callback

    
class ReviewDataApp:
    def __init__(self):
        """
        Object that renders specified components and communicates with a ReviewData object to modify
        ReviewData object's annotation and history.

        """
        self.more_components = OrderedDict()
        
    def columns_to_string(self, df, columns):
        new_df = df.copy()
        for c in columns:
            new_df[c] = new_df[c].astype(str)

        return new_df
        
    def run(
        self,
        review_data: ReviewDataInterface,
        annot_app_display_types_dict: Dict = None,
        autofill_dict: Dict = None,
        review_data_table_df: pd.DataFrame = None,
        review_data_table_page_size: int = 10,
        collapsable=True,
        auto_export: bool = True,
        auto_export_path: Union[Path, str] = None, 
        attributes_to_export: List = ['annot_df', 'history_df'],
        mode='external',
        host='0.0.0.0',
        port=8050,
        hide_history_df_cols=[],
        components_name_order=[],
    ):

        """
        Run the app

        Parameters
        ----------
        review_data: ReviewDataInterface
            ReviewData object to review with the app

        mode: {'inline', 'external', 'tab'}
            How to display the dashboard

        host: str
            Host address

        port: int
            Port access number

        annot_app_display_types_dict: Dict
            at run time, determines how the inputs for annotations will be displayed

            Format: {annot_name: app_display_type}

            Rules:
                - annot_name: must be a key in ReviewData.data.annot_col_config_dict
                - app_display_type: see valid_annotation_app_display_types

        autofill_dict: Dict
            At run time, adds buttons for each component included and if clicked, defines how to autofill
            annotations for the review_data object using the specified data corresponding to the component
            
            Format: {button_name_1: {annot_name: State()},
                     button_name_2: {annot_name: State(), another_annot_name: 'a literal value'}}
            
            Rules:
                - button_name:  must be a name of a button
                - annot_name:      must be the name of an annotation type in the review_data (ReviewData) data
                                   annot_col_config_dict
                - autofill values: State()'s referring to objects in the component named component name, or a 
                                   valid literal value according to the DataAnnotation object's validation method.
                                   
        auto_export: bool, default=False
            Whether to auto export on save to path set by argument auto_export_path
            
        auto_export_path: Union[Path, str]
            Path to export data if argument auto_export=True
            
        attributes_to_export: List
            List of attributes from the data object to automatically export

        hide_history_df_cols: List[str]
            list of columns in the history table to NOT display in the dashboard

        components_name_order: List[str]
            list of component names in order to be displayed. Component names not included will not be displayed.
            Be careful if some components listen to each other. To list component names:

            > reviewer.app.more_components
        """
        multi_type_columns = [c for c in annot_app_display_types_dict.keys() if review_data.data.annot_col_config_dict[c].annot_value_type == 'multi']

        self.history_display_cols = review_data.data.history_df.columns
        self.history_display_cols = [c for c in self.history_display_cols if c not in hide_history_df_cols]

        if len(components_name_order) > 0:
            self.ordered_more_components = {name: self.more_components[name] for name in components_name_order}
        else:
            self.ordered_more_components = self.more_components

        def get_history_display_table(subject_index_value):
            filtered_history_df = review_data.data.history_df.loc[
                review_data.data.history_df['index'] == subject_index_value,
                self.history_display_cols
            ].loc[::-1]
            
            return self.columns_to_string(filtered_history_df, multi_type_columns)
        
        
        # app = JupyterDash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])
        app = Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])
        
        reviewed_data_df = pd.DataFrame(index=review_data.data.index, columns=['label'])
        reviewed_data_df['label'] = reviewed_data_df.apply(lambda r: self.gen_dropdown_labels(review_data, r), 
                                                           axis=1)
        reviewed_data_df.index.name = 'value'
        
        app.layout, annotation_panel_component, autofill_buttons, gen_autofill_states, autofill_literals = \
            self.gen_layout(
                review_data,
                reviewed_data_df,
                annot_app_display_types_dict,
                autofill_dict,
                review_data_table_df=review_data_table_df,
                review_data_table_page_size=review_data_table_page_size,
                collapsable=collapsable,
                multi_type_columns=multi_type_columns
            )
        
        # data_pkl_fn = data_path/data.pkl where data_path is the path to data that is set in set_review_data
        # e.g. if data_path = './reviewer_data' then data_pkl_fn = './reviewer_data/data.pkl' and app.title = 'reviewer_data'
        app.title = review_data.data_pkl_fn.split('/')[-2] 

        if auto_export:
            if auto_export_path is None:
                auto_export_path = f"{review_data.data_pkl_fn.rsplit('.', 1)[0]}.auto_export"
                print(f'Setting auto_export_path to {auto_export_path}')
            if not os.path.exists(auto_export_path):
                print(f'Making directory {auto_export_path} for auto exporting.')
                os.mkdir(auto_export_path)
            
            print(f'Using {auto_export_path} for auto exporting.')
        
        def validate_callback_outputs(
            component_output, 
            component, 
            which_callback='callback function(s)'
        ):
            if not isinstance(component_output, list) or (len(component.callback_output) != len(component_output)):
                raise ValueError(
                    f'Component ({component.name}) {which_callback} does not return '
                     'the same length output as component\'s callback_output.\n'
                     'Make sure your output from the callback function is a list, '
                     'and the values correspond to your defined callback_outputs list\n'
                     f'Expected output: {component.callback_output}\n'
                     f'Runtime callback output: {component_output}\n'
                )
            return

        more_component_inputs = {
            c.name: c.callback_input + c.callback_state + c.callback_state_external for c_name, c in self.ordered_more_components.items()
        }

        more_component_inputs_as_states = {
            c_name: [
                State(**{'component_id': c.component_id, 'component_property': c.component_property}) for c in c_list
            ] for c_name, c_list in more_component_inputs.items()
        }

        more_component_outputs = {c.name: c.callback_output for c_name, c in self.ordered_more_components.items()}
               
        def update_components(output_dict, subject_index_value, more_component_inputs_as_states):
            output_dict['more_component_outputs'] = {
                c.name: list(np.full(len(c.callback_output), dash.no_update)) for c_name, c in self.ordered_more_components.items()
            }
            
            for c_name, component in self.ordered_more_components.items():
                if component.new_data_callback is not None:
                    component_output = component.new_data_callback(
                        review_data.data,
                        subject_index_value,
                        *more_component_inputs_as_states[component.name]
                    )
                    validate_callback_outputs(component_output, component, which_callback='new_data_callback')
                    output_dict['more_component_outputs'][component.name] = component_output
            
            history_df = get_history_display_table(subject_index_value)
            output_dict['history_table'] = history_df.to_dict('records')
            output_dict['history_table_selected_row_state'] = []
            
            if history_df.empty:
                output_dict['annot_panel'] = {
                    annot_name: annot_display_component.default_display_value if \
                    annot_display_component.default_display_value is not None else '' \
                    for annot_name, annot_display_component in annot_app_display_types_dict.items() 
                }
            else:
                current_annotations = review_data.data.annot_df.loc[subject_index_value, list(annot_app_display_types_dict.keys())].to_dict()
                for annot_name, v in current_annotations.items():
                    if isinstance(v, list):
                        continue
                    
                    if (
                        (pd.isna(v) or (v == '') or (v is None)) and 
                        (annot_app_display_types_dict[annot_name].default_display_value is not None)
                    ):
                        current_annotations[annot_name] = annot_app_display_types_dict[annot_name].default_display_value
                    
                output_dict['annot_panel'] = current_annotations
            return output_dict

        ###### Callbacks
        @app.callback(
            output=dict(
                is_open=Output('APP-freeze-modal', 'is_open')
            ),
            inputs=dict(
                freeze_button=Input('APP-freeze-button', 'n_clicks'),
                freeze_close=Input('APP-freeze-close', 'n_clicks'),
                freeze_confirm=Input('APP-freeze-confirm', 'n_clicks')
            ),
            prevent_initial_call=True,
        )
        def toggle_freeze_modal(freeze_button, freeze_close, freeze_confirm):
            if ctx.triggered_id=='APP-freeze-button':
                return {'is_open': True}
            return {'is_open': False}
        
        @app.callback(
            output=dict(
                subscript=Output('APP-title-subscript', 'children'),
                subscript_color=Output('APP-title-subscript', 'style'),
                button=Output('APP-freeze-button', 'children'),
                history_display_table=Output('APP-history-table', 'data', allow_duplicate=True),
                dropdown_options=Output('APP-dropdown-data-state', 'options', allow_duplicate=True),
                annot_panel=annotation_panel_component.callback_output
            ),
            inputs=dict(
                freeze_confirm=Input('APP-freeze-confirm', 'n_clicks'),
                annotations_confirm=State('APP-annotations-checkbox', 'value'),
                history_display_table=State('APP-history-table', 'data'),
                dropdown_options=State('APP-dropdown-data-state', 'options'),
                annot_panel=annotation_panel_component.callback_state 
            ),
            prevent_initial_call=True,
        )
        def freeze_data(freeze_confirm, annotations_confirm, history_display_table, dropdown_options, annot_panel):
            subscript = '  TEST MODE'
            subscript_color = {'color': 'red'}
            button = dbc.Button('Freeze Data', id='APP-freeze-button')

            if freeze_confirm:
                if not annotations_confirm:
                    # remove data
                    annot_cols = review_data.data.annot_df.columns.tolist()
                    review_data.data.annot_df = pd.DataFrame(index=review_data.data.index, columns=annot_cols)
                    review_data.data.history_df = pd.DataFrame(columns=['index', 'timestamp', 'source_data_fn']+annot_cols)
                    review_data.save_data()

                    # removed displayed data
                    history_display_table = pd.DataFrame().to_dict('records')
                    reviewed_data_df['label'] = reviewed_data_df.index
                    dropdown_options = reviewed_data_df.reset_index().to_dict('records')
                    annot_panel = {annot_name: '' for annot_name in annot_app_display_types_dict.keys()}
                    
                review_data.mh.set_attribute('freeze_data', True)
                review_data.mh.set_attribute('freeze_data_timestamp', datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

            if review_data.mh.metadata['freeze_data']:
                subscript = f'  Frozen at {review_data.mh.metadata["freeze_data_timestamp"]}'
                subscript_color = {'color': 'black'}
                button = html.Div([])

            return {
                'subscript': subscript, 
                'subscript_color': subscript_color, 
                'button': button, 
                'history_display_table': history_display_table,
                'dropdown_options': dropdown_options, 
                'annot_panel': annot_panel
            }

        @app.callback(
            output=dict(
                more_component_outputs=more_component_outputs,
                history_table=Output('APP-history-table', 'data', allow_duplicate=True),
                history_table_selected_row_state=Output('APP-history-table', 'selected_rows', allow_duplicate=True),
                annot_panel=annotation_panel_component.callback_output,
                review_data_selected_value=Output('APP-review-data-table', 'selected_rows', allow_duplicate=True),
                review_data_page_current=Output('APP-review-data-table', 'page_current', allow_duplicate=True)
            ),
            inputs=dict(
                dropdown_value=Input('APP-dropdown-data-state', 'value'),
                review_data_table_state=State('APP-review-data-table', 'data'),
                more_component_inputs_as_states=more_component_inputs_as_states,
                review_data_table_derived_virtual_data_state=State('APP-review-data-table', 'derived_virtual_data')
            ),
            prevent_initial_call=True,
        )
        def update_sample_via_dropdown(
            dropdown_value,
            review_data_table_state,
            more_component_inputs_as_states,
            review_data_table_derived_virtual_data_state
        ):
            """
            Update the whole dashboard with the corresponding data of 
            the selected subject in the dropdown menu
            """
            output_dict = {
                'review_data_page_current': dash.no_update,
                'review_data_selected_value': dash.no_update,
            }
            
            tmp_review_data_table_df = pd.DataFrame.from_records(review_data_table_state)
            subject_index_value = dropdown_value

            if review_data_table_df is not None: # if the table is being used
                index_tmp_review_data_table_df = tmp_review_data_table_df.loc[
                    tmp_review_data_table_df['index'] == subject_index_value
                ]
                output_dict['review_data_selected_value'] = index_tmp_review_data_table_df.index

                review_data_table_derived_virtual_df = pd.DataFrame.from_records(
                    review_data_table_derived_virtual_data_state)
                index_relative_review_data_table_df = review_data_table_derived_virtual_df.loc[
                    review_data_table_derived_virtual_df['index'] == subject_index_value,
                ]
                if not index_relative_review_data_table_df.empty:
                    output_dict['review_data_page_current'] = floor(index_relative_review_data_table_df.index[0] / review_data_table_page_size)


            output_dict = update_components(output_dict, subject_index_value, more_component_inputs_as_states)
            return output_dict
            

        @app.callback(
            output=dict(
                dropdown_value=Output('APP-dropdown-data-state', 'value', allow_duplicate=True),
                more_component_outputs=more_component_outputs,
                history_table=Output('APP-history-table', 'data', allow_duplicate=True),
                history_table_selected_row_state=Output('APP-history-table', 'selected_rows', allow_duplicate=True),
                annot_panel=annotation_panel_component.callback_output,
            ),
            inputs=dict(
                review_data_selected_value=Input('APP-review-data-table', 'selected_rows'),
                review_data_table_state=State('APP-review-data-table', 'data'),
                more_component_inputs_as_states=more_component_inputs_as_states,
            ),
            prevent_initial_call=True,
        )
        def update_sample_via_review_table(
            review_data_selected_value,
            review_data_table_state,
            more_component_inputs_as_states
        ):
            """
            Update the whole dashboard with the corresponding data of the selected subject in the dash table
            """
            output_dict = {}
            
            tmp_review_data_table_df = pd.DataFrame.from_records(review_data_table_state)
            subject_index_value = tmp_review_data_table_df.loc[review_data_selected_value[0], 'index'] 
            output_dict['dropdown_value'] = subject_index_value

            output_dict = update_components(output_dict, subject_index_value, more_component_inputs_as_states)
            return output_dict
            

        @app.callback(
            output=dict(
                history_table=Output('APP-history-table', 'data', allow_duplicate=True),
                dropdown_list_options=Output('APP-dropdown-data-state', 'options', allow_duplicate=True),
                review_data_table_data=Output('APP-review-data-table', 'data', allow_duplicate=True),
                test_mode_alert=Output('APP-test-mode-alert', 'is_open')
            ),
            inputs=dict(
                submit_annot_button=Input('APP-submit-button-state', 'n_clicks'),
                annot_input_state=annotation_panel_component.callback_state,
                dropdown_value=State('APP-dropdown-data-state', 'value'),
                review_data_table_state=State('APP-review-data-table', 'data'),
            ),
            prevent_initial_call=True,
        )
        def submit_button_annotation(
            submit_annot_button,
            annot_input_state,
            dropdown_value,
            review_data_table_state
        ):
            """
            Save current annotations to the annot_df field, and update the dropdown menu timestamp and history table
            """
            output_dict = {'review_data_table_data': dash.no_update, 'test_mode_alert': False}
            if not review_data.mh.metadata['freeze_data']:
                output_dict['test_mode_alert'] = True
            for annot_name in annot_app_display_types_dict.keys():
                annot_type = review_data.data.annot_col_config_dict[annot_name]
                # Convert type
                if annot_app_display_types_dict[annot_name].display_output_format is not None:
                    annot_input_state[annot_name] = annot_app_display_types_dict[annot_name].display_output_format(
                        annot_input_state[annot_name]
                    )
                
                validate_annot_data(annot_type, annot_input_state[annot_name])
                    
                new_annot_input_state = dict(annot_input_state)
                review_data._update(dropdown_value, new_annot_input_state)
                
                if auto_export:
                    review_data.export_data(auto_export_path, attributes_to_export=attributes_to_export, verbose=False)

                output_dict['history_table'] = get_history_display_table(dropdown_value).to_dict('records')
                
                reviewed_data_df.loc[dropdown_value, 'label'] = self.gen_dropdown_labels(
                    review_data,
                    reviewed_data_df.loc[dropdown_value]
                )
                output_dict['dropdown_list_options'] = reviewed_data_df.reset_index().to_dict('records')

                if review_data_table_df is not None:
                    tmp_review_data_table_df = pd.DataFrame.from_records(review_data_table_state).set_index('index')
                    tmp_review_data_table_df.loc[
                        dropdown_value,
                        review_data.data.annot_df.columns
                    ] = review_data.data.annot_df.loc[dropdown_value].values
                    output_dict['review_data_table_data'] = self.columns_to_string(
                        tmp_review_data_table_df, 
                        multi_type_columns
                    ).reset_index().to_dict('records')

                return output_dict

               
        @app.callback(
            output=dict(
                annot_panel=annotation_panel_component.callback_output
            ),
            inputs=dict(
                autofill_buttons=[Input(b.id, 'n_clicks') for b in autofill_buttons],
                autofill_states=gen_autofill_states,
            ),
            prevent_initial_call=True,
        )
        def autofill_annot(
            autofill_buttons,
            autofill_states
        ):
            """
            Detect and fill annotation panel from autofill trigger
            """
            ctx = dash.callback_context
            if not ctx.triggered:
                raise PreventUpdate
            else:
                prop_id = ctx.triggered[0]['prop_id'].split('.')[0]
                
            output_dict = {
                'annot_panel': {
                    annot_col: dash.no_update for annot_col in annot_app_display_types_dict.keys()
                }
            }
            component_name = prop_id.split('APP-autofill-')[-1]
            for autofill_annot_col, value in autofill_states[prop_id].items():
                output_dict['annot_panel'][autofill_annot_col] = value
                
            for autofill_annot_col, value in autofill_literals[prop_id].items():
                output_dict['annot_panel'][autofill_annot_col] = value
            return output_dict


        @app.callback(
            output=dict(
                annot_panel=annotation_panel_component.callback_output
            ),
            inputs=dict(
                revert_annot_button=Input('APP-revert-annot-button', 'n_clicks'),
                history_table_selected_row_state=State('APP-history-table', 'selected_rows'),
                dropdown_value=State('APP-dropdown-data-state', 'value')
            ),
            prevent_initial_call=True,
        )
        def revert_annot(
            revert_annot_button,
            history_table_selected_row_state,
            dropdown_value
        ):
            """
            Revert the annotation fields to a previous annotation by selecting the row in the history table
            """
            output_dict = {
                'annot_panel': {
                    annot_col: dash.no_update for annot_col in annot_app_display_types_dict.keys()
                }
            }
            if len(history_table_selected_row_state) > 0:
                history_df = review_data.data.history_df.loc[review_data.data.history_df['index'] == dropdown_value].loc[::-1].reset_index()
                output_dict['annot_panel'] = history_df.iloc[
                    history_table_selected_row_state[0]
                ][review_data.data.annot_df.columns].fillna('').to_dict()
                
            return output_dict


        @app.callback(
            output=dict(
                annot_panel=annotation_panel_component.callback_output
            ),
            inputs=dict(
                clear_annot_button=Input('APP-clear-annot-button', 'n_clicks'),
            ),
            prevent_initial_call=True,
        )
        def clear_annot(clear_annot_button):
            """
            Clear the data currently entered in the annotation panel
            """
            output_dict = {}
            output_dict['annot_panel'] = {annot_col: '' for annot_col in review_data.data.annot_df.columns}
            return output_dict

               
        @app.callback(
            output=dict(
                more_component_outputs=more_component_outputs
            ),
            inputs=dict(
                more_component_inputs=more_component_inputs,
                dropdown_value=State('APP-dropdown-data-state', 'value'),
            ),
            prevent_initial_call=True,
        )
        def internal_update_components(more_component_inputs, dropdown_value):
            """
            Update triggered component
            """
            ctx = dash.callback_context
            if not ctx.triggered:
                raise PreventUpdate
            else:
                prop_id = ctx.triggered[0]['prop_id'].split('.')[0]

            output_dict = {
                'more_component_outputs': {
                    c.name: list(np.full(len(c.callback_output), dash.no_update)) for c_name, c in self.ordered_more_components.items()
                }
            }
            for c_name, component in self.ordered_more_components.items():
                if sum([ci.component_id == prop_id for ci in component.callback_input]) > 0:
                    if component.internal_callback is None:
                        raise ValueError(
                            f'Component ({component.name}) has Inputs that change ({prop_id}), '
                            f'but no internal_callback defined to handle it.'
                            f'Either remove Input "{prop_id}" from "{component.name}.callback_input" attribute, '
                            f'or define a callback function'
                        )
                    if dropdown_value:
                        component_output = component.internal_callback(
                            review_data.data,
                            dropdown_value,
                            *more_component_inputs[component.name]
                        )

                        validate_callback_outputs(component_output, component, which_callback='internal_callback')
                        output_dict['more_component_outputs'][component.name] = component_output
            return output_dict
        
        if not review_data.mh.metadata['freeze_data']:
            warnings.warn(
                'You are in test mode. Your data will not be saved.'
            )

        jupyter_dash.default_mode = mode
        app.run(host=host, port=port, debug=True)
        
    def gen_layout(
        self,
        review_data: ReviewDataInterface,
        reviewed_data_df: pd.DataFrame,
        annot_app_display_types_dict: Dict,
        autofill_dict: Dict,
        review_data_table_df: pd.DataFrame=None,
        review_data_table_page_size: int = 10,
        collapsable=True,
        multi_type_columns=[]
    ):
        """
        Generate layout of the dashboard
        """
        review_data_title = html.Div([
            dbc.Alert('This is a test mode. You must freeze your data to save your annotations.', color='danger', id='APP-test-mode-alert', is_open=False, duration=4000),
            html.H1(review_data.data_pkl_fn.split('/')[-2], style={'display': 'inline'}),
            dbc.Modal([
                dbc.ModalHeader("Are you sure you want to freeze the data?"),
                dbc.ModalBody("Input data will not be changable once data is frozen."),
                html.Div([
                    dbc.Checkbox(id='APP-annotations-checkbox', label='Save current annotations', value=True)
                ], style={'margin-left': '15px'}),
                dbc.ModalFooter([
                    dbc.Button("Freeze data", id="APP-freeze-confirm", n_clicks=0, className="ml-auto"),
                    dbc.Button("Cancel, I am still testing", id="APP-freeze-close", n_clicks=0, className="ml-auto")
                ]),
            ], id="APP-freeze-modal"),
            html.Div(children=html.Button('Freeze data') if not review_data.mh.metadata['freeze_data'] else html.Div([]), id='APP-freeze-button')
        ])
        review_data_path = html.Div([html.P(f'Path: {review_data.data_pkl_fn}')])
        review_data_description = html.Div([html.P(f'Description: {review_data.data.description}')])
        dropdown = html.Div(dcc.Dropdown(options=reviewed_data_df.reset_index().to_dict('records'),
                                         value=None, 
                                         id='APP-dropdown-data-state'))
        
        dropdown_component = AppComponent(name='APP-dropdown-component',
                                          layout=[dropdown], 
                                          use_name_as_title=False)

        if review_data_table_df is not None:
            if set(review_data_table_df.index.tolist()) != set(reviewed_data_df.index.tolist()):
                raise ValueError(
                    f'{review_data_table_df.index} has missing or extra index values. '
                    f'Index values should look like: {reviewed_data_df.index.tolist()}'
                )
            new_review_data_table_df = review_data_table_df.copy()
            new_review_data_table_df.index.name = 'index'
            new_review_data_table_df = pd.concat(
                [
                    new_review_data_table_df, 
                    self.columns_to_string(review_data.data.annot_df, multi_type_columns)
                ], 
                axis=1
            )
            review_data_table_data = new_review_data_table_df.reset_index().to_dict('records')
            review_data_table_columns = new_review_data_table_df.reset_index().columns.tolist()
            style = {'display': 'block'}

        else:
            review_data_table_layout = html.Div(html.H1('None'), style={'display': 'none'})
            review_data_table_data = []
            review_data_table_columns = []
            style = {'display': 'none'}

        review_data_table_layout = html.Div(
            dash.dash_table.DataTable(
                id='APP-review-data-table',
                data=review_data_table_data,
                columns=[
                    {"name": i, "id": i, "deletable": False, "selectable": False} for i in review_data_table_columns
                ],
                editable=False,
                filter_action="native",
                sort_action="native",
                sort_mode="multi",
                column_selectable="single",
                row_selectable="single",
                selected_columns=[],
                selected_rows=[],
                page_action="native",
                page_current=0,
                page_size=review_data_table_page_size,
                style_data={
                        'whiteSpace': 'normal',
                        'height': 'auto',
                },
            ),
            style=style,
        )

        review_data_table_component = AppComponent(
            name='APP-review-data-table-app-component',
            layout=review_data_table_layout,
            use_name_as_title=False,
        )


        history_table = html.Div([
            html.H2('History Table'),
            html.Button(
                'Revert to selected annotation',
                id=f'APP-revert-annot-button',
                n_clicks=0,
                style={"marginBottom": "15px"}),
            html.Button(
                'Clear annotation inputs',
                id=f'APP-clear-annot-button',
                n_clicks=0,
                style={"marginBottom": "15px"}),
            html.Div(
                dash.dash_table.DataTable(
                    id='APP-history-table',
                    data=pd.DataFrame().to_dict('records'),
                    columns=[
                        {"name": i, "id": i, "deletable": False, "selectable": False} for i in self.history_display_cols
                    ],
                    editable=False,
                    filter_action="native",
                    sort_action="native",
                    sort_mode="multi",
                    column_selectable="single",
                    row_selectable="single",
                    selected_columns=[],
                    selected_rows=[],
                    page_action="native",
                    page_current=0,
                    page_size=5,
                    style_data={
                        'whiteSpace': 'normal',
                        'height': 'auto',
                    }
                ),
                style={"maxHeight": "400px", "overflow": "scroll"},
            )
        ])
        
        history_component = AppComponent(name='APP-history-component',
                                         layout=[history_table], 
                                         use_name_as_title=False)
        
        autofill_buttons, autofill_states, autofill_literals = self.gen_autofill_buttons_and_states(review_data,
                                                                                                    autofill_dict)

        annotation_panel_component = self.gen_annotation_panel_component(review_data, autofill_buttons,
                                                                         annot_app_display_types_dict)
        
        
        if collapsable:
            more_components_layout = dbc.Accordion(
                [dbc.AccordionItem(c.layout, title=c_name) for c_name, c in self.ordered_more_components.items()], 
                always_open=True, 
                start_collapsed=False,
                active_item=[f'item-{i}' for i in range(len(self.ordered_more_components.keys()))],
                style={'.accordion-button': {'font-size': 'xx-large'}}
            )
        else:
            more_components_layout = [dbc.Row(c.layout, style={"marginBottom": "15px"}) for c_name, c in self.ordered_more_components.items()]

        
        layout = html.Div(
            [
                dbc.Row([review_data_title], style={"marginBottom": "15px"}),
                dbc.Row([review_data_path], style={"marginBottom": "15px"}),
                dbc.Row([review_data_description], style={"marginBottom": "15px"}),
                dbc.Row([dropdown_component.layout], style={"marginBottom": "15px"}),
                dbc.Row([review_data_table_component.layout], style={"marginBottom": "15px"}),
                dbc.Row([dbc.Col(annotation_panel_component.layout, width=5),
                        dbc.Col(html.Div(history_component.layout), width=7)], 
                       style={"marginBottom": "15px"}),
                dbc.Row(more_components_layout)
            ],
             style={'marginBottom': 50, 'marginTop': 25, 'marginRight': 25, 'marginLeft': 25})
        
        
        all_ids = np.array(get_component_ids(layout))
        check_duplicates(all_ids, 'full component')

        return layout, annotation_panel_component, autofill_buttons, autofill_states, autofill_literals

    def gen_dropdown_labels(self, review_data: ReviewDataInterface, r: pd.Series):
        data_history_df = review_data.data.history_df[review_data.data.history_df["index"] == r.name]
        if not data_history_df.empty:
            return str(r.name) + f' (Last update: {data_history_df["timestamp"].tolist()[-1]})'
        else:
            return r.name
       
    def gen_autofill_buttons_and_states(self, review_data, autofill_dict):
        review_data_annot_names = list(review_data.data.annot_col_config_dict.keys())
            
        autofill_buttons = []
        autofill_states = {}
        autofill_literals = {}
        for autofill_button_name, button_autofill_dict in autofill_dict.items():

            # check keys in component_autofill_dict are in ReviewData annot columns
            annot_keys = np.array(list(button_autofill_dict.keys()))
            missing_annot_refs = np.array([annot_k not in review_data_annot_names for annot_k in annot_keys])
            if missing_annot_refs.any():
                raise ValueError(
                    f'Reference to annotation columns {annot_keys[np.argwhere(missing_annot_refs).flatten()]}'
                    f' do not exist in the Review Data Object.'
                    f'Available annotation columns are {review_data_annot_names}')

            # check states exist in layout
            button_autofill_states_dict = {annot_type: state for annot_type, state in button_autofill_dict.items()
                                              if isinstance(state, State)}
            button_autofill_state_ids = np.array(
                get_callback_io_ids(list(button_autofill_states_dict.values()), expected_io_type=State))

            all_ids = get_component_ids([c.layout for c_name, c in self.more_components.items()])
            missing_autofill_state_ids = [i for i in button_autofill_state_ids if i not in all_ids]
            if len(missing_autofill_state_ids) > 0:
                raise ValueError(f'State ids do not exist anywhere in the layout: '
                                 f'{missing_autofill_state_ids}\n')

            # check literal fill values are valid
            button_autofill_non_states_dict = {annot_type: item for annot_type, item in
                                               button_autofill_dict.items() if not isinstance(item, State)}

            for annot_type, non_state_value in button_autofill_non_states_dict.items():
                if isinstance(non_state_value, Input) or isinstance(non_state_value, Output):
                    raise ValueError(
                        f'Invalid autofill object {non_state_value} for annotation column "{annot_type}". '
                        'Either pass in a State() object or literal value')

                try:
                    review_data.data.annot_col_config_dict[annot_type].validate(non_state_value)
                except ValueError:
                    raise ValueError(
                        f'Autofill value for annotation column '
                        f'"{annot_type}" failed validation. '
                        f'Check validation for annotation column "{annot_type}": \n'
                        f'{review_data.data.annot_col_config_dict[annot_type]}')

            # Make button    
            autofill_button_component = html.Button(autofill_button_name,
                                                    id=f'APP-autofill-{autofill_button_name}',
                                                    n_clicks=0,
                                                    style={"marginBottom": "15px"})
            autofill_buttons += [autofill_button_component]
            autofill_states[autofill_button_component.id] = button_autofill_states_dict
            autofill_literals[autofill_button_component.id] = button_autofill_non_states_dict

        if len(autofill_buttons) == 0:
            autofill_button_component = html.Button(
                'dummy button - required for autofill callback',
                id=f'APP-autofill-dummy-button',
                disabled=True, hidden=True,
                style={"marginBottom": "15px"}
            )
            autofill_buttons += [autofill_button_component]
            
        return autofill_buttons, autofill_states, autofill_literals


    def gen_annotation_panel_component(
        self,
        review_data: ReviewDataInterface,
        autofill_buttons: [html.Button],
        annot_app_display_types_dict: Dict
    ):

        if len(annot_app_display_types_dict) == 0:
            raise ValueError(
                f'annot_app_display_types_dict is empty. '
                f'Make sure to run reviewer.set_default_review_data_annotations_configuration(), '
                f'or manually set up your annotations'
            )
        submit_annot_button = html.Div([
            html.Button(id='APP-submit-button-state', 
                n_clicks=0, 
                children='Submit', 
                style={"marginBottom": "15px"}
            ),
            html.Sub(
                '  TEST MODE', 
                style={'color': 'red'}, 
                id='APP-title-subscript'
            ) if not review_data.mh.metadata['freeze_data'] else html.Sub(
                f'  Frozen at {review_data.mh.metadata["freeze_data_timestamp"]}', 
                style={'color': 'black'}, 
                id='APP-title-subscript'
            ),
        ])
        
        def annotation_display_component_input(annot_name, annot, annot_app_display_type):
            input_component_id = f"APP-{annot_name}-{annot_app_display_type}-input-state"
            
            input_component = annot_app_display_type.gen_input_component(
                annot, 
                component_id=input_component_id, 
            )
                
            return dbc.Row([dbc.Label(annot_name, html_for=input_component_id, width=2),
                            dbc.Col(input_component)],
                           style={"margin-bottom": "15px"})
        
        panel_components = autofill_buttons + [
            annotation_display_component_input(
                annot_name, review_data.data.annot_col_config_dict[annot_name], display_type
            ) for annot_name, display_type in annot_app_display_types_dict.items()
        ] + [submit_annot_button]
        
        panel_inputs = [Input('APP-submit-button-state', 'nclicks')]
        return AppComponent(
            name='APP-Panel',
            layout=panel_components,
            callback_output={
                annot_name: Output(f"APP-{annot_name}-{display_type}-input-state", "value", allow_duplicate=True)
                for annot_name, display_type in annot_app_display_types_dict.items()},
            callback_input=panel_inputs,
            callback_state={
                annot_name: State(f"APP-{annot_name}-{display_type}-input-state", "value") for
                    annot_name, display_type in annot_app_display_types_dict.items()},
            use_name_as_title=False
        )
    
    
    def add_component(
        self, 
        component: AppComponent, 
        **kwargs
    ):
        """
        component: An AppComponent object to include in the app
        **kwargs: include more arguments for the component's callback functions 
        """
        all_component_names = [c_name for c_name, c in self.more_components.items()]
        
        if component.name in all_component_names:
            warnings.warn(f'There is a component named "{component.name}" already in the app. '
                          f'Change the name of the component to add it to the current layout')
            return
        
        ids_with_reserved_prefix_list = np.array(['APP-' in i for i in component.all_component_ids])
        if ids_with_reserved_prefix_list.any():
            raise ValueError(f'Some ids use reserved keyword "APP-". Please change the id name\n'
                             f'Invalid component ids: '
                             f'{component.all_component_ids[np.argwhere(ids_with_reserved_prefix_list).flatten()]}')
        
        new_component = copy.deepcopy(component)
        if component.new_data_callback is not None:
            new_component.new_data_callback = lambda *args: component.new_data_callback(*args, **kwargs)
        else:
            new_component.new_data_callback = None
            
        if component.internal_callback is not None:
            new_component.internal_callback = lambda *args: component.internal_callback(*args, **kwargs)
        else:
            new_component.internal_callback = None

        self.more_components[component.name] = new_component
        all_ids = get_component_ids([c.layout for c_name, c in self.more_components.items()])
        check_duplicates(all_ids, f'ids found in previously added component from component named {component.name}')
        
    def add_table_from_path(self, data_table_source, table_title, component_id, table_fn_col, table_cols):
        """
        Parameters
        ----------
        data_table_source: attribute name of the dataframe to use from the Data object
        table_title:     Title of the table
        component_id: component name for the table
        table_fn_col:   column in review_data data dataframe with file path with table to display
        table_cols:     columns to display in table from table_fn_col
        """
        table = html.Div(dbc.Table.from_dataframe(pd.DataFrame()),
                         id=component_id)
        self.add_component(
            AppComponent(
                table_title,
                table,
                new_data_callback=lambda data, idx: [dbc.Table.from_dataframe(
                    pd.read_csv(getattr(data, data_table_source).loc[idx, table_fn_col],
                                sep='\t',
                                encoding='iso-8859-1')[table_cols])],
                callback_output=[Output(component_id, 'children')]
            )
        )


def get_component_ids(component: Union[List, Tuple]):
    if isinstance(component, list) or isinstance(component, tuple):
        id_list = []
        for comp in component:
            id_list += get_component_ids(comp)
        return id_list
    elif isinstance(component, str) or isinstance(component, int) or isinstance(component, float):
        return []
    elif 'children' in component.__dict__.keys(): 
        children_ids = get_component_ids(component.children)
        if 'id' in component.__dict__.keys():
            children_ids += [component.id]
        return children_ids
    else:
        return [component.id] if 'id' in component.__dict__.keys() else []


def get_callback_io_ids(io_list, expected_io_type):
    
    if isinstance(io_list, list):
        items = io_list
    elif isinstance(io_list, dict):
        items = [item for k, item in io_list.items()]
    else:
        raise ValueError(f'Input list is not a list or dictionary: {io_list}')

    wrong_io_type = np.array([not isinstance(item, expected_io_type) for item in items])
    if wrong_io_type.any():
        raise ValueError(f'Some or all items in list are not the expected type {expected_io_type}: {items}')
        
    return [item.component_id for item in items]


def check_callback_io_id_in_list(ids, all_ids, ids_type: str, all_ids_type: str):
    ids_not_in_all_ids = np.array(ids)[np.argwhere(np.array([id_name not in all_ids for id_name in ids])).flatten()]
    if len(ids_not_in_all_ids) > 0:
        raise ValueError(f"{ids_type} ids not in {all_ids_type}:\n"
                         f"{ids_type} ids: {ids_not_in_all_ids}\n"
                         f"{all_ids_type} ids: {all_ids}")


def check_duplicates(a_list, list_type: str):
    values, counts = np.unique(a_list, return_counts=True)
    if (counts > 1).any():
        raise ValueError(f"Duplicate {list_type}: {values[np.argwhere(counts > 1)].flatten()}")


def check_duplicate_objects(a_list, list_type: str):
    check_duplicates([c.__str__() for c in a_list], list_type=list_type)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="AnnoMate",
    version="1.0.2",
    author="Claudia Chu",
    author_email="cchu@broadinstitute.org",
    description="A general tool to create dashboards for manual review",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/getzlab/AnnoMate",
    project_urls={
        "Bug Tracker": "https://github.com/getzlab/AnnoMate/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "."},
    packages=setuptools.find_packages(where=".", exclude=['zeption.examples', '']),
    python_requires=">=3.8", # last tested version: 3.9
    install_requires = ['setuptools>=70.0.0', # security vulnerability with versions up to 69.1.1 
                        # 'setuptools<60.0', # https://numpy.org/doc/stable/reference/distutils_status_migration.html. Install before numpy
                        'cnv-suite>=0.2.0',
                        'dash>=2.11.0',
                        'dash-bootstrap-components',
                        'dash-cytoscape',
                        'dash-daq',
                        'fsspec',
                        'gcsfs',
                        'google-auth',
                        'google-api-core',
                        'hound',
                        'ipykernel',
                        'ipython',
                        'jupyterlab>=4.0.2',
                        'matplotlib',
                        'pandas',
                        'pickleshare',
                        'pillow',
                        'pip',
                        'plotly>=5.15.0',
                        'scipy',
                        'frozendict',
                        # fixes jupyter-dash bug when repeat calls to run_server hangs
                        'flask>=3.0.3',
                        'werkzeug>=2.3.3', # security vulnerability with 2.2.2
                       ]
)   




from AnnoMate.Data import Data
import pandas as pd
from typing import Dict


class PatientSampleData(Data):

    def __init__(self,
                 index,
                 description,
                 participant_df: pd.DataFrame,
                 sample_df: pd.DataFrame,
                 annot_df: pd.DataFrame = None,
                 annot_col_config_dict: Dict = None,
                 history_df: pd.DataFrame = None,
                 ):
        super().__init__(index=index,
                         description=description,
                         annot_df=annot_df,
                         annot_col_config_dict=annot_col_config_dict,
                         history_df=history_df)

        self.sample_df = sample_df
        self.participant_df = participant_df

        # check tables are linked, other validation
        self.participant_col_in_sample = 'participant'
